#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <sys/errno.h>

#include <mad.h>

#define DTMFX_IMPLEMENTATION 1
#include "decode_mad.h"
#include "decoder.h"
#include "util.h"

#ifndef BUFSIZ
/* setbuf(3) indicates that BUFSIZ (when declared) is optimal */
#define BUFSIZ 1024
#endif

#include <mad.h>

// This file draws originally (but by now distantly) on the libmad example file minimad.c

typedef struct {
    unsigned char *buffer;
    size_t buffer_ngood;        /* number of good bytes in buffer (may be < buffer_length) */
    size_t buffer_length;       /* allocated size */
    int buffer_finished_p;      /* flag indicating eof */
    int fd;                     /* non-negative if we're reading from an FD */
    /* nbufferfuls is the number of times the input function has been
       called (if we're reading from an FD) */
    int nbufferfuls;
    enum mad_flow (*input_callback)(void* data, struct mad_stream *stream);
    void *analyser_state;       /* state of caller, to be passed back to process_frame */
} MadInputHandler;

static enum mad_flow input_from_mmap(void* data, struct mad_stream *stream);
static enum mad_flow input_from_fd(void* data, struct mad_stream *stream);
static MadInputHandler* read_from_file(const char *filename);
static MadInputHandler* read_from_fd(const int fd);
static int decode_mp3_internal(void*, MadInputHandler*);
static void release_input_handler(MadInputHandler*);

/*
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded. In this example, an entire file
 * has been mapped into memory, so we just call mad_stream_buffer() with the
 * address and length of the mapping. When this callback is called a second
 * time, we are finished decoding.
 */
static enum mad_flow input_from_mmap(void* data, struct mad_stream *stream)
{
    MadInputHandler *mih = (MadInputHandler*)data;

    if (mih->buffer_finished_p) {
        return MAD_FLOW_STOP;
    }

    // one-shot call with entire pre-existing buffer
    mad_stream_buffer(stream, mih->buffer, mih->buffer_length);
    mih->buffer_finished_p = 1;

    return MAD_FLOW_CONTINUE;
}

/*
 * An input callback for reading from a file-descriptor.  This is
 * slightly intricate, since we have to re-use the portion of input
 * data which is still unused but lying in the buffer.  See the
 * (libmad distribution's) madplay example application, in particular
 * player.c:decode_input_read, which accounts for a large part of the
 * API 'documentation'.
 */
static enum mad_flow input_from_fd(void* data, struct mad_stream *stream)
{
    MadInputHandler *mih = (MadInputHandler*)data;
    size_t nread;

    if (mih->buffer_finished_p) {
        return MAD_FLOW_STOP;
    }

    if (stream->next_frame) {
        // find the number of bytes in the buffer which haven't been
        // read, because they're in an incomplete frame
        size_t input_length = &mih->buffer[mih->buffer_ngood] - stream->next_frame;
        assert(input_length > 0);
        // move these to the front of the buffer
        memmove(mih->buffer, stream->next_frame, input_length);
        // we have this many good bytes so far...
        mih->buffer_ngood = input_length;
    }

    assert(0 <= mih->buffer_ngood && mih->buffer_ngood < mih->buffer_length);

    do {
        nread = read(mih->fd,
                     mih->buffer + mih->buffer_ngood, 
                     mih->buffer_length - mih->buffer_ngood);
    }
    while (nread == -1 && errno == EINTR);

    if (nread == -1) {
        error_exit("Error reading from MP3 input stream: %s", strerror(errno));
    } else if (nread == 0) {
        // set a flag for next time
        mih->buffer_finished_p = 1;
    }
    // note how many good bytes we have, for next time
    mih->buffer_ngood += nread;
    assert(0 < mih->buffer_ngood && mih->buffer_ngood <= mih->buffer_length);
    mih->nbufferfuls++;
    mad_stream_buffer(stream, mih->buffer, mih->buffer_ngood);

    return MAD_FLOW_CONTINUE;
}

static MadInputHandler* read_from_file(const char *filename)
{
    size_t filesize;
    MadInputHandler* handler;

    if ((handler = (MadInputHandler*)malloc(sizeof(MadInputHandler))) == NULL) {
        error_exit("Can't allocate %ld for MadInputHandler", sizeof(MadInputHandler));
    }
    
    /* initialize our private state structure */
    handler->buffer = (unsigned char*)map_file(filename, &filesize);
    handler->buffer_length = filesize;
    handler->buffer_ngood = filesize;

    handler->input_callback = &input_from_mmap;

    handler->fd = -1;              /* negative, to indicate we're reading from map */
    handler->nbufferfuls = -1;     /* not used */
    handler->buffer_finished_p = 0;

    return handler;
}

static MadInputHandler* read_from_fd(const int fd)
{
    MadInputHandler* handler;

    assert(fd >= 0);
    
    if ((handler = (MadInputHandler*)malloc(sizeof(MadInputHandler))) == NULL) {
        error_exit("Can't allocate %ld for MadInputHandler", sizeof(MadInputHandler));
    }

    /* initialize our private state structure */
    handler->buffer_length = BUFSIZ;
    handler->buffer_ngood = 0;     /* none read yet */
    handler->buffer  = (unsigned char*)malloc(handler->buffer_length);
    if (handler->buffer == NULL) {
        error_exit("Unable to allocate read buffer in mad_read_from_fd");
    }

    handler->fd = fd;              /* non-negative */
    handler->nbufferfuls = 0;
    handler->input_callback = &input_from_fd;
    handler->buffer_finished_p = 0;

    return handler;
}

static void release_input_handler(MadInputHandler *h)
{
    if (h->fd >= 0) {
        free(h->buffer);
        h->buffer = NULL;
        h->fd = -1;
    } else if (h->buffer != NULL) {
        // mapped file
        unmap_file(h->buffer, h->buffer_length);
        h->buffer = NULL;
    }
    
    free(h);
}

/*
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */
static enum mad_flow error_handler(void *data,
                                   struct mad_stream *stream,
                                   struct mad_frame *frame)
{
    MadInputHandler *input_handler = (MadInputHandler*)data;
    
    if (stream->error == MAD_ERROR_LOSTSYNC) {
        if (input_handler->nbufferfuls > 1) {
            // this is near the beginning or end of the file, so we can presume
            // it's hitting ID3 metadata; so ignore it silently
        } else {
            // ignore it noisily
            fprintf(stderr, "%s: lost sync: ignored\n", program_name());
        }
    } else {
        fprintf(stderr, "%s: decoding error 0x%04x (%s)\n",
                program_name(),
                stream->error, mad_stream_errorstr(stream));
        return MAD_FLOW_BREAK;
    }

    return MAD_FLOW_CONTINUE;
}

static enum mad_flow process_mad_frame(void *data,
                                       struct mad_header const *header,
                                       struct mad_pcm *pcm)
{
    mad_fixed_t const *ch;
    int samplerate = pcm->samplerate;
    int nsamples = pcm->length;

    MadInputHandler *input_handler = (MadInputHandler*)data;

    static Sample *sample_buf = NULL;
    Sample* p;
    Sample* end_p;
    static size_t sample_bufsize = 0;

    if (sample_bufsize < nsamples) {
        if (sample_buf != NULL) {
            free(sample_buf);
        }
        sample_bufsize = nsamples;
        if ((sample_buf = (Sample*)malloc(sample_bufsize * sizeof(Sample))) == NULL) {
            error_exit("Can't allocate memory for %ld floats", sample_bufsize);
        }
    }

    /* If there happens to be more than one channel, ignore all but the first */
    ch = pcm->samples[0];

    end_p = &sample_buf[nsamples];
    for (p=sample_buf; p<end_p; p++, ch++) {
        *p = mad_f_todouble(*ch); /* there isn't a mad_f_tofloat */
    }

    process_frame(input_handler->analyser_state,
                  sample_buf,
                  samplerate,
                  nsamples);

    return MAD_FLOW_CONTINUE;
}

/*
 * This is the function which does the libmad decoding.
 * It instantiates a decoder object and configures it with the input,
 * output, and error callback functions above. A single call to
 * mad_decoder_run() continues until a callback function returns
 * MAD_FLOW_STOP (to stop decoding) or MAD_FLOW_BREAK (to stop decoding and
 * signal an error).
 */
static int decode_mp3_internal(void* analyser_state, MadInputHandler* input_handler)
{
    struct mad_decoder decoder;
    int decoder_status;

    input_handler->analyser_state = analyser_state;
    
    /* configure input, output, and error functions */
    mad_decoder_init(&decoder,
                     input_handler,
                     input_handler->input_callback,
                     0 /* header */,
                     0 /* filter */,
                     process_mad_frame,
                     error_handler,
                     0 /* message */);

    /* start decoding */
    decoder_status = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

    /* release the decoder */
    mad_decoder_finish(&decoder);

    return decoder_status;
}

/*
 * This function is responsible for decoding the input file, and
 * handing the contents to the decoder function
 * decoder:process_frame.  We do this by setting up callbacks
 * appropriately in decode_mp3_internal.
 * If argument fd >= 0, then we read from this (open) fd; if not, then
 * we read from the file named by 'filename'.
 */
int decode_mp3_with_libmad(const int fd, const char* filename, void* analyser_state)
{
    MadInputHandler* mih = NULL;
    int status;

    if (fd >= 0) {
        mih = read_from_fd(fd);
    } else if (filename != NULL) {
        mih = read_from_file(filename);
    } else {
        error_exit("Bad call to decode_mp3_with_libmad: no filename or fd");
    }
    assert(mih != NULL);

    status = decode_mp3_internal(analyser_state, mih);

    release_input_handler(mih);

    return status;
}
