#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <sys/errno.h>
#include <math.h>

#define DTMFX_IMPLEMENTATION 1
#include "decoder.h"
#include "decode_mad.h"
#include "read_wav.h"
#include "goertzel.h"
#include "util.h"

#ifndef BUFSIZ
/* setbuf(3) indicates that BUFSIZ (when declared) is optimal */
#define BUFSIZ 1024
#endif

typedef enum {
    IN_GAP, POSSIBLE_TONE, DEFINITE_TONE, POSSIBLE_GAP, AFTER_DIGITS 
} ToneState;

#define NFREQ 8
typedef struct {
    // input buffer
    //MadInputHandler* mih;

    // details of the filter
    float filter_length;        /* in seconds */
    int filter_nsamples;
    int filter_samp_per_sec;    /* negative if uninitialised, 0 if to be reinitialised */
    // ...which are encapsulated in...
    GoertzelFilter f[NFREQ];

    // manage the identification of DTMF pulses
    ToneState tone_state;
    int current_tone;           /* current candidate tone */
    int n_possible_tones;       /* number of times this tone has been spotted recently */
    double min_tone_length;     /* cutoff: tones shorter than this are rejected */
    double min_power;           /* cutoff: bands with power below this are rejected */
    float noise_floor;          /* smallest signal regarded as reasonable */

    // accumulate statistical information from the scan of the stream
    float offset_s;             /* current offset into recording (seconds) */
    float dtmf_start;           /* start of first tone (seconds) */
    float dtmf_last;            /* start of the most recent tone to be identified (seconds) */
    float dtmf_last_candidate;  /* start of most recent candidate tone */
    double sum_gap;             /* accumulating sum of gaps between digits */
    double sum_gap2;            /* accumulating sum of (gaps between digits)^2 */
    int analyse_stats_p;        /* true if we are to analyse statistics */

    // the list of hex digits read from the stream
    unsigned char* result;
    int result_allocated;
    int ndigits;
} FilterState;


static DTMFX_Result* decode_audio_file(const int fd, const char* filename,
                                       FilterState *state,
                                       DTMFX_FilterParameters*,
                                       DTMFX_FileFormat format);

/* Initialise FilterParameter block to default/sentinel values */
void dtmfx_initialise_filter_parameters(DTMFX_FilterParameters* p)
{
    p->filter_length = -1;
    p->background_mult = -1;
    p->min_tone_length = -1;
    p->collect_all_digits_p = 0;

    return;
}

/* Decode the contents of the named file, or stdin if filename==NULL.
 * If force_format is not FILE_UNKNOWN, then it forces the
 * interpretation of the input file.
 */
DTMFX_Result* dtmfx_decode_bytes_from_audio(const char* filename,
                                            DTMFX_FilterParameters *params,
                                            DTMFX_FileFormat force_format)
{
    DTMFX_Result *result;
    FilterState state;
    int fnlen;
    DTMFX_FileFormat fmt = FILE_MP3;  /* default */

    int fd;

    if (filename == NULL) {
        fd = STDIN_FILENO;
    } else {
        /* sentinel value for decode_audio_file: read from the file
         * named in the argument 'filename'
         */
        fd = -1;
        
        fnlen = strlen(filename);
        if (strcmp(&filename[fnlen-4], ".wav") == 0) {
            fmt = FILE_WAV;
        }
    }
    if (force_format != FILE_UNKNOWN) {
        fmt = force_format;
    }
    
    assert(fd >= 0 || filename != NULL);

    result = decode_audio_file(fd, filename, &state, params, fmt);
    
    return result;
}

/*
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to output (or play) the decoded PCM audio.
 */
int process_frame(void *data,
                  Sample* samples, /* a frame of samples -- not multiplexed */
                  int samplerate,  /* samples/sec */
                  int nsamples)    /* number of samples in the frame */
{
    static int frame_count = 0;  /* used only for debugging/chatter messages */
    static int sample_count = 0; /* used only for debugging/chatter messages */
    double secs_per_sample;      /* in current frame */
    int sample_no;               /* index */
    int highcount;               /* the number of times a signal has to
                                  * be above threshold before we deem
                                  * it to be above threshold in fact */

    FilterState *state = (FilterState*)data;

    if (state->filter_samp_per_sec > 0 && state->filter_samp_per_sec != samplerate) {
        int i;
        if (program_verbosity() > 0) {
            fprintf(stderr, "decoder: Changing from %d samp/sec to %d samp/sec\n",
                    state->filter_samp_per_sec, samplerate);
        }
        for (i=0; i<NFREQ; i++) {
            goertzel_uninit(&state->f[i]);
        }
        state->filter_samp_per_sec = 0; /* force goertzel_init */
    }
    if (state->filter_samp_per_sec <= 0) { /* initialise or reinitialise */
        int i;
        double tmp;
        int samples_in_filter = state->filter_length * samplerate;

        // re-initialise for new frame
        state->filter_nsamples = samples_in_filter;
        state->filter_samp_per_sec = samplerate;
        for (i=0; i<NFREQ; i++) {
            goertzel_init(&state->f[i],
                          (i<4 ? dtmf_low_freq(i) : dtmf_high_freq(i-4)),
                          samplerate,
                          state->filter_length);
        }

        /*
         * We need a minimum cut-off for the power/digits calculation,
         * since we don't want to be confused by tiny fluctuations
         * near the noise floor.  You would guess that a good way of
         * doing this is to gather statistics from the first few tens
         * of ms at the start of the audio, but it seems that some
         * encoders put near-silence here (or libmad extracts
         * near-silence here, which comes to the same thing), so we
         * can't get any useful data.
         *
         * Instead, estimate the minimum as being the power for amplitude noise_floor.
         *
         *** An estimate of $g_{n,l}$
         ** Obtain an estimate of the value of $g_{n,l}$, given
         ** $x$.
         **
         ** Write $x(t)=A\cos\bar\omega t$, where
         ** $\bar\omega$ is frequency in radians/sec.  Then
         ** \begin{align*}
         ** g(T) &= \int_0^T x(t)\exp i\bar\omega t\,dt\\
         **      &= A\int_0^T(\cos^2\bar\omega t + i\cos\bar\omega t\sin\bar\omega t)\,dt\\
         **      &= \frac{A}{\bar\omega} \frac{\bar\omega T}\pi \frac{\pi}{2}
         **       = \frac{AT}{2},
         ** \end{align*}
         ** if $\bar\omega T$ is a multiple of $\pi$, since $\int_0^\pi\cos z\sin z\,dz = 0$
         ** and $\int_0^\pi\cos^2z\,dz=\pi/2$.
         ** If $\bar\omega T$ is not a multiple of $\pi$, then the
         ** imaginary component picks up only the excess over a
         ** multiple of $\pi$, which decreases as a fraction of the
         ** real part, as $T$ increases.
         **
         ** Another route to this is to write $\Delta t=T/n$ and
         ** $\omega=\bar\omega\Delta t/2\pi$, the frequency in
         ** radians/sample.  Then
         ** \begin{align*}
         ** g(T)\approx g_n(T) &= A \sum_k^n\cos^2(\bar\omega k\Delta t)\cdot\Delta t\\
         **        &= A\Delta t\sum_k^n \cos^2\bar\omega k\Delta t\\
         **        &= \frac{AT}{n} \sum^n \cos^2 2\pi\omega k\\
         **        &= \frac{AT}{2}
         ** \end{align*}
         ** since the average value of $\cos^2x$ is 1/2.  Now,
         ** comparing this with Eqn.(\ref{e:gnl}), we can see that
         ** $g_n(T) \sim ng_{n,l}/T$, so that an
         ** estimate of $g_{n,l}$ is $An/2$, if $\omega$ is a resonance.
         *
         * An estimate of g_{n,l} is An/2, so the estimated power is this squared.
         */
        tmp = samples_in_filter * state->noise_floor / 2.0;
        state->min_power = tmp*tmp;
    }

    secs_per_sample = 1.0/samplerate;
    /* The following is the number of samples in 5ms; this is
     * heuristic.  A better value would depend on an estimated
     * variability of the power signal.
     */ 
    highcount = 0.005 * samplerate;

    assert(state->filter_samp_per_sec > 0);
        
    if (program_verbosity() > 0) {
        fprintf(stderr, "frame: %d@%d, %d samples (highcount %d, min %g)\n",
                ++frame_count, sample_count, nsamples,
                highcount, state->min_power);
    }

    /*
     * We want to work out the threshold for a signal being 'high',
     * and we want to make this dynamically dependent on the inputs,
     * so we don't have to worry about picking numbers.
     *
     * We don't know the amplitude of the incoming signal, so pick a
     * threshold based on the power in the various bands.  Get the
     * average power in the low and the high bands, and recognise a
     * signal if there's one which is more than double this value, and
     * three which are less than it.
     */
    for (sample_no=0; sample_no<nsamples; sample_no++) {
        int i;
        Sample sampf;
        double high_threshold = 0;
        unsigned char flag_high = 0;
        int chatter_p, debug_p;
        double g[8];
        
        sample_count++;
        state->offset_s += secs_per_sample;

        if (state->tone_state == AFTER_DIGITS) {
            continue;
        }

        sampf = samples[sample_no];
        
        chatter_p = (program_verbosity() > 0);
        debug_p = (program_verbosity() > 1);
        
        for (i=0; i<NFREQ; i++) {
            g[i] = goertzel_step(&state->f[i], sampf);
            if (debug_p) {
                fprintf(stderr, "%g ", g[i]);
            }

            high_threshold += g[i];
        }

        /*
         * The average of the eight powers is high_threshold/8;
         * high_threshold/4 is therefore twice the average.
         */
        high_threshold /= 4;
        // Cut this off at min_power
        if (high_threshold < state->min_power) {
            high_threshold = state->min_power;
        }
        if (debug_p) {
            fprintf(stderr, "%g %g\n", sampf, high_threshold);
        }

        // spot the bands which have power above the threshold
        for (i=0; i<NFREQ; i++) {
            if (g[i] > high_threshold) {
                flag_high |= 1<<i;
            }
        }

        /*
         * Now check to see if precisely one of the first four, and
         * one of the second four, bits is set.
         */
        {
            unsigned char d;
            switch (flag_high) {
              case 0x11: d = 0x1; break;
              case 0x12: d = 0x4; break;
              case 0x14: d = 0x7; break;
              case 0x18: d = 0xE; break;
              case 0x21: d = 0x2; break;
              case 0x22: d = 0x5; break;
              case 0x24: d = 0x8; break;
              case 0x28: d = 0x0; break;
              case 0x41: d = 0x3; break;
              case 0x42: d = 0x6; break;
              case 0x44: d = 0x9; break;
              case 0x48: d = 0xF; break;
              case 0x81: d = 0xA; break;
              case 0x82: d = 0xB; break;
              case 0x84: d = 0xC; break;
              case 0x88: d = 0xD; break;
              default:   d = DTMF_NONE; break;
            }


            /*
             * State machine.  There are five states, given by
             * ToneState, and with the following internal states
             *
             *   in_gap
             *   possible_tone(digit, ntones)
             *   definite_tone(digit)
             *   possible_gap(digit, ntones)
             *   after_digits
             *
             * The transitions are controlled by whether a tone
             * corresponding to digit d is detected:
             *
             * in_gap:
             *         tone?               -> possible_tone(digit, 1)
             *         not tone?           -> in_gap
             *
             * possible_tone(digit, ntones):
             *         tone = digit?       => tmp=ntones+1
             *         not tone = digit?   => tmp=ntones-1
             *         ...
             *         "gap to last tone too long"
             *                             -> after_digits
             *         tmp = HIGHCOUNT     -> definite_tone(digit)
             *         tmp = 0             -> in_gap
             *         otherwise           -> possible_tone(digit, tmp)
             *
             * definite_tone(digit):
             *         tone = digit?       -> definite_tone(digit)
             *         not tone = digit?   -> possible_gap(digit, HIGHCOUNT-1)
             *
             * possible_gap(digit, ntones):
             *         tone = digit?       => tmp=ntones+1
             *         not tone = digit?   => tmp=ntones-1
             *         ...
             *         tmp = HIGHCOUNT     -> definite_tone(digit)
             *         tmp = 0             -> in_gap
             *         otherwise           -> possible_gap(digit, tmp)
             *
             * after_digits:
             *                             -> after_digits        
             */
            switch (state->tone_state) {
              case IN_GAP:
                if (d != DTMF_NONE) {
                    // this might be a tone
                    state->current_tone = d;
                    state->n_possible_tones = 1;
                    // Set the time of the tone start to be in the
                    // middle of the interval covered by the filter.
                    // This presumes (I think) that the filter length
                    // is just long enough to detect the power in this
                    // band; this may not be sensible.
                    state->dtmf_last_candidate = state->offset_s - state->filter_length/2;
                    state->tone_state = POSSIBLE_TONE;
                }
                break;

              case POSSIBLE_TONE:
                assert (0 < state->n_possible_tones && state->n_possible_tones < highcount);
                if (d == state->current_tone) {
                    state->n_possible_tones++;
                } else {
                    state->n_possible_tones--;
                }
                assert (0 <= state->n_possible_tones && state->n_possible_tones <= highcount);
                if (state->n_possible_tones == highcount) {
                    // we've found a digit!

                    if (chatter_p) {
                        // for debugging purposes, report the exact
                        // sample where we decide this, rather than
                        // adjusting it to sample_count-(state->filter_nsamples/2)
                        fprintf(stderr, "digit: %x=%x (@%d/%g)\n", flag_high, d,
                                sample_count, state->dtmf_last_candidate);
                    }
                    if (state->result == NULL) {
                        state->result_allocated = 4;
                        if ((state->result = (unsigned char*)malloc(state->result_allocated)) == NULL) {
                            error_exit("Can't allocate chars for result");
                        }
                    } else if (state->ndigits == state->result_allocated) {
                        state->result_allocated *= 2;
                        if ((state->result = (unsigned char*)realloc(state->result, state->result_allocated)) == NULL) {
                            error_exit("Unable to realloc chars for result");
                        }
                    }
                    assert(state->result != NULL && state->ndigits < state->result_allocated);

                    if (state->ndigits == 0) {
                        // first digit
                        state->dtmf_start = state->dtmf_last_candidate;
                    } else if (state->analyse_stats_p) {
                        /*
                         * Consistency check:
                         *
                         * Examine the gaps between digits, and calculate
                         * the mean and standard deviation of those gaps.
                         * If the difference between the most recent gap,
                         * and the mean gap, is more than a few stddev,
                         * then deem this digit, and all following it, to
                         * be spurious, and stop accumulating them.
                         *
                         * At this point we have already seen state->ndigits digits,
                         * so there is one less than that number of gaps.
                         * Because of the way we detect digits by their
                         * prominence above a background, and because the
                         * background starts off low, the first digit is
                         * detected earlier in its rise than the
                         * subsequent ones, so we initially overestimate
                         * the gap between digits.
                         */
                        int ngaps = state->ndigits - 1;
                        double gap = state->dtmf_last_candidate - state->dtmf_last;
                        if (ngaps > 4) { // let the mean and stddev settle before drawing conclusions
                            double mean = state->sum_gap/ngaps;
                            double stddev = sqrt(state->sum_gap2/ngaps - mean*mean);
                            double norm_dev = (gap-mean)/stddev; // normalised deviation
                            if (chatter_p) {
                                fprintf(stderr,
                                        "gap at digit %d  mean=%g  stddev=%g:  gap=%g  dgap=%.3fsd\n",
                                        state->ndigits + 1, mean, stddev, gap, norm_dev);
                            }
                            if (abs(norm_dev) > 5.0) {
                                // 5 is heuristic, but 5-sigma is the
                                // traditional 'detection' threshold.
                                // We can afford to be generous in this.
                                if (chatter_p) {
                                    fprintf(stderr, "excess gap deviation: stopping accumulating digits\n");
                                }
                                state->tone_state = AFTER_DIGITS;
                                continue; /* JUMP OUT and continue with the outer loop */
                            }
                        }
                        state->sum_gap += gap;
                        state->sum_gap2 += gap*gap;
                    }
                    state->dtmf_last = state->dtmf_last_candidate;
                    state->result[state->ndigits] = d;
                    state->ndigits++;

                    state->tone_state = DEFINITE_TONE;
                } else if (state->n_possible_tones == 0) {
                    state->tone_state = IN_GAP;
                } else {
                    assert (0 < state->n_possible_tones && state->n_possible_tones < highcount);
                    // no transition
                }
                break;

              case DEFINITE_TONE:
                if (d != state->current_tone) {
                    state->n_possible_tones = highcount-1;
                    state->tone_state = POSSIBLE_GAP;
                }
                // else no transition
                break;

              case POSSIBLE_GAP:
                assert (0 < state->n_possible_tones && state->n_possible_tones < highcount);
                if (d == state->current_tone) {
                    state->n_possible_tones++;
                } else {
                    state->n_possible_tones--;
                }
                assert (0 <= state->n_possible_tones && state->n_possible_tones <= highcount);
                if (state->n_possible_tones == highcount) {
                    // false alarm -- this wasn't a gap after all
                    state->tone_state = DEFINITE_TONE;
                } else if (state->n_possible_tones == 0) {
                    // end of the tone
                    float tone_end_time = state->offset_s - state->filter_length/2;
                    if (tone_end_time - state->dtmf_last < state->min_tone_length) {
                        if (chatter_p) {
                            fprintf(stderr, "digit end (@%d/%g); length (%g < %g) too short; ignored\n",
                                    sample_count, tone_end_time,
                                    tone_end_time - state->dtmf_last, state->min_tone_length);
                        }
                        state->ndigits--;
                    } else if (debug_p) {
                        fprintf(stderr, "digit end (@%d/%g); length %g\n",
                                sample_count, tone_end_time,
                                tone_end_time - state->dtmf_last);
                    }
                    state->tone_state = IN_GAP;
                } else {
                    assert (0 < state->n_possible_tones && state->n_possible_tones < highcount);
                    // no transition
                }
                break;

              default:
                error_exit("!Shouldn't happen: impossible tone state: %d", state->tone_state);
            }
        }
    }

    return 0;
}

/*
 * This is the function which does all the decoding.  It sets up
 * filter state, and then calls the decode_mp3_with_libmad function.
 * If argument fd >= 0, then we read from this (open) fd; if not, then
 * we read from the file named by 'filename'.
 */
static DTMFX_Result* decode_audio_file(const int fd,
                                       const char* filename,
                                       FilterState* state,
                                       DTMFX_FilterParameters* params,
                                       DTMFX_FileFormat format)
{
    int decoder_status;
    DTMFX_Result *result;
    
    // Initialise various filter state values, which are common to both callers of this function
    state->filter_length = (params->filter_length < 0 ? 0.02 : params->filter_length);
    state->filter_samp_per_sec = -1; /* meaning: completely uninitialised */
    state->noise_floor = 0.025 * (params->background_mult < 0 ? 1.0 : params->background_mult);
    state->result = NULL;
    state->offset_s = 0;
    state->dtmf_start = -1;      /* negative sentinel value */
    state->dtmf_last = -1;
    state->sum_gap = 0;
    state->sum_gap2 = 0;
    state->result = NULL;
    state->ndigits = 0;
    state->tone_state = IN_GAP;
    /* enforce 30ms minimum tone length -- heuristic */
    state->min_tone_length = (params->min_tone_length < 0 ? 0.030 : params->min_tone_length);
    state->analyse_stats_p = ! params->collect_all_digits_p;
    state->min_power = 0;

    switch (format) {
      case FILE_MP3:
        decoder_status = decode_mp3_with_libmad(fd, filename, state);
        break;
        
      case FILE_WAV:
        if (fd >= 0) {
            error_exit("Can't (yet) read WAV files from stdin");
        }
        decoder_status = read_audio_wav(-1, filename, state);
        break;

      default:
        error_exit("!Impossible file format %d", format);
    }

    if ((result = (DTMFX_Result*)malloc(sizeof(DTMFX_Result))) == NULL) {
        error_exit("Can't allocate space for decoder result");
    }

    result->decoded = state->result;
    result->decoded_length = state->ndigits;
    result->start_offset = state->dtmf_start;
    result->audio_length = state->offset_s;

    // don't free state->result, becaue it's now owned by *result
    
    return result;
}

void dtmfx_decoder_free_result(DTMFX_Result* result)
{
    if (result->decoded != NULL) {
        free(result->decoded);
    }
    free(result);
}
