#ifndef DTMF_H
#define DTMF_H 1

#include <stdint.h>

/*
 * DTMF constants
 * See ITU Q.23 <http://www.itu.int/rec/T-REC-Q.23/en>
 * and <http://en.wikipedia.org/wiki/DTMF>.
 *
 * See dtmf.c for functions dtmf_low_freq() and dtmf_high_freq().
 *
 * ITU Q.24 <http://www.itu.int/rec/T-REC-Q.24/en> includes
 * recommended minima for (amongst other things) pulse and gap durations.
 */

/**
 * Information about the DTMF-encoded audio to be generated.
 */
typedef struct {
    int bits_per_sample;        /* -4 for IEEE-754 float */
    int bytes_per_sample;       /* (bits_per_sample+7)/8 */
    int sample_rate;            /* samples per second */
    int nchannels;
    float duration;             /* seconds */
    float gap_duration;         /* seconds */
    float start_padding;        /* seconds */
    float end_padding;          /* seconds */
    float volume;               /* volume as a fraction of maximum */
} DTMFX_SoundInfo;

typedef uint8_t byte;

/* 
 * The following definitions encode in a single byte the grid of low and
 * high frequencies.  The index of the low frequency is in the low
 * pair of bits, and that of the high frequency in the high pair.
 * Thus '2', which is low-frequency 0 and high frequency 1, is 0x40
 *
 *        1209  1336  1477  1633
 *   697   1=0   2=4   3=8   a=c
 *   770   4=1   5=5   6=9   b=d
 *   852   7=2   8=6   9=a   c=e
 *   941   e=3   0=7   f=b   d=f
 *
 * 'E'='*' and 'F'='#'
 */
#define DTMF_0 0x7
#define DTMF_1 0x0
#define DTMF_2 0x4
#define DTMF_3 0x8
#define DTMF_4 0x1
#define DTMF_5 0x5
#define DTMF_6 0x9
#define DTMF_7 0x2
#define DTMF_8 0x6
#define DTMF_9 0xa
#define DTMF_A 0xc
#define DTMF_B 0xd
#define DTMF_C 0xe
#define DTMF_D 0xf
#define DTMF_E 0x3
#define DTMF_F 0xb
#define DTMF_NONE 0xff

#if DTMFX_IMPLEMENTATION

int dtmf_low_freq(int);
int dtmf_high_freq(int);

#endif /* DTMFX_IMPLEMENTATION */


#endif /* DTMF_H 1 */
