#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include <config.h>

#include "dtmfx.h"
#include "jason/src/jason.h"

static void Usage(int);
static char* progname;


int main(int argc, char** argv)
{
    const char* audio_filename = NULL;
    int exit_status = 0;
    DTMFX_Result *decoding;
    int ch;
    int verbosity = DTMFX_QUIET;
    DTMFX_FileFormat input_format = FILE_UNKNOWN;
    enum { FMT_STRING, FMT_JSTRING, FMT_JSON, FMT_JJSON } output_format = FMT_STRING;
    DTMFX_FilterParameters filter_params;

    dtmfx_initialise_filter_parameters(&filter_params);

    progname = argv[0];         /* for reporting below */
    dtmfx_set_program_name(argv[0]);

    while ((ch = getopt(argc, argv, "ghf:l:t:T:V")) != -1) {
        switch (ch) {
          case 'f':
            /*
             * Input format (the type converted 'f'rom: "mp3" or "wav".
             * (In fact, this option is almost redundant, since it is
             * mainly _intended_ for the case where we are reading a
             * .wav file from stdin, and so can't detect the type from
             * the filename extension.  But read_wav.c:read_audio_wav
             * can't read WAVs from stdin, yet).
             */
            if (strcmp(optarg, "mp3") == 0) {
                input_format = FILE_MP3;
            } else if (strcmp(optarg, "wav") == 0) {
                input_format = FILE_WAV;
            } else {
                Usage(1);
            }
            break;
          case 't':
            /*
             * to format: 'json' or 'string'
             */
            if (strcmp(optarg, "string") == 0) {
                output_format = FMT_STRING;
            } else if (strcmp(optarg, "jstring") == 0) {
                output_format = FMT_JSTRING;
            } else if (strcmp(optarg, "json") == 0) {
                output_format = FMT_JSON;
            } else if (strcmp(optarg, "jjson") == 0) {
                output_format = FMT_JJSON;
            } else {
                Usage(1);
            }
            break;
          case 'g':
            if (verbosity < DTMFX_DEBUG) {
                verbosity++;
            }
            dtmfx_set_program_verbosity(verbosity);
            break;
          case 'l':
            /*
             * length of filter.
             * If the figure is <= 1.0, it is in seconds;
             * or if > 1.0, it is in ms.
             */
            {
                float n = strtof(optarg, NULL);
                if (n <= 0.0) {
                    error_exit("Can't have negative filter length: %f", n);
                } else if (n <= 1.0) {
                    filter_params.filter_length = n;
                } else {
                    filter_params.filter_length = n/1000.0;
                }
                break;
            }
          case 'T':
            /*
             * Tuning parameters
             */
            {
                char *tune_arg = strchr(optarg, '=');
                int keylen;
                if (tune_arg == NULL) {
                    error_exit("Malformed tuning parameter");
                }
                keylen = tune_arg - optarg;
                tune_arg++;
                if (strncmp(optarg, "threshold", keylen) == 0) {
                    /*
                     * Background multiplier: we ignore signals below
                     * a certain floor, on the grounds that they're
                     * probably spurious; this allows us to tune this
                     * number: if this is 1.0, the required multiplier is the default.
                     */
                    float n = strtof(tune_arg, NULL);
                    if (n <= 0.0) {
                        error_exit("Threshold multiplier must be positive (not %g)", n);
                    }
                    filter_params.background_mult = n;
                } else if (strncmp(optarg, "all-digits", keylen) == 0) {
                    if (*tune_arg == '1') {
                        filter_params.collect_all_digits_p = 1;
                    } else {
                        error_exit("all-digits argument must be '1'");
                    }
                } else if (strncmp(optarg, "min-tone-length", keylen) == 0) {
                    float n = strtof(tune_arg, NULL);
                    if (n <= 0) {
                        error_exit("Minimum tone length must be positive (not %g)", n);
                    }
                    if (n > 1) {
                        n /= 1000; /* convert from ms to seconds */
                    }
                    filter_params.min_tone_length = n;
                } else {
                    error_exit("Unrecognised tuning parameter");
                }
                break;
            }
          case 'V':
            {
                const char* snapshotid = SNAPSHOTID;
                if (snapshotid[0] == '\0') {
                    // release version
                    printf("This is dtmf2hex.\nFrom %s v%s, released %s\nSee <https://bitbucket.org/nxg/dtmfx>\n",
                           PACKAGE_NAME, PACKAGE_VERSION, RELEASEDATE);
                } else {
                    printf("This is dtmf2hex.\nFrom %s v%s, snapshot %s\n",
                           PACKAGE_NAME, PACKAGE_VERSION, snapshotid);
                }
                exit(0);
            }
            break;
          case 'h':
          case '?':
            Usage(0);
          default:
            Usage(1);
        }
    }
    argc -= optind;
    argv += optind;

    switch (argc) {
      case 0:
        audio_filename = NULL;
        break;
      case 1:
        audio_filename = argv[0];
        break;
      default:
        Usage(1);
        break;
    }

    decoding = dtmfx_decode_bytes_from_audio(audio_filename,
                                             &filter_params,
                                             input_format);
    if (decoding == NULL || decoding->decoded == NULL) {
        fprintf(stderr, "%s: unable to detect any DTMF tones in the file %s\n",
                progname, (audio_filename == NULL ? "<stdin>" : audio_filename));
        exit_status = 1;
    } else {
        byte* p = decoding->decoded;
        int i;

        if (output_format == FMT_JSTRING) {
            printf("{ \"string\": \"");
            for (i=0; i<decoding->decoded_length; i++, p++) {
                printf("%x", *p);
            }
            printf("\", \"start_offset\": %.3f, \"audio_length\": %.3f, \"end_offset\": %.3f }\n",
                   decoding->start_offset, decoding->audio_length,
                   decoding->audio_length - decoding->start_offset);
        } else if (output_format == FMT_STRING) {
            for (i=0; i<decoding->decoded_length; i++, p++) {
                printf("%x", *p);
            }
            printf("\n");
        } else if (output_format == FMT_JSON || output_format == FMT_JJSON) {
            byte* b;
            byte *nibbles = decoding->decoded;
            JsonObject obj;
            size_t nbytes = decoding->decoded_length;

            // convert the nibbles to bytes, and decode them to JSON
            if (nbytes % 2 == 1) {
                fprintf(stderr, "Decoded an odd number of bytes -- can't decode to JSON\n");
                exit(1);
            }
            
            nbytes /= 2;
            if ((b = (byte*)malloc(nbytes)) == NULL) {
                fprintf(stderr, "Can't allocate %ld bytes for decoding\n", nbytes);
                exit(1);
            }
            for (i=0; i<nbytes; i++) {
                b[i] = nibbles[2*i] * 0x10 + nibbles[2*i+1];
            }

            obj = parse_der_bytes(b, nbytes, NULL);
            if (obj != NULL) {
                const char* json_s = print_json_object(obj);
                if (output_format == FMT_JSON) {
                    puts(json_s);
                } else {
                    printf("{ \"object\": %s, "
                           "\"start_offset\": %.3f, "
                           "\"audio_length\": %.3f, "
                           "\"end_offset\": %.3f }\n",
                           json_s,
                           decoding->start_offset, decoding->audio_length,
                           decoding->audio_length - decoding->start_offset);
                }
            } else {
                fprintf(stderr, "Couldn't parse bytes as JSON\n");
                exit_status = 1;
            }
            free(b);
        } else {
            error_exit("!Unrecognised output_format: %d", output_format);
        }

        dtmfx_decoder_free_result(decoding);
    }

    return exit_status;
}

static void Usage(int exit_status)
{
    fprintf(stderr, "Usage: %s [-ghV] [-fhex] [-t format] [-l length] [file.mp3]\n", progname);
    exit(exit_status);
}
