/*
 * Implementation of a Goertzel filter.
 * See <http://en.wikipedia.org/wiki/Goertzel_algorithm>
 */
/*
*** Definition of the Goertzel filter
**
** Given a sequence of input samples $x_i$, We can define a convolution
** of this with a complex exponential, via:
** \begin{equation}
**   g_{n,l} = \sum_{k=l-n+1}^{l} x_k \Omega^k,
** \label{e:gnl}
** \end{equation}
** where $\Omega=\exp 2\pi i\omega$, and $\omega$ is expressed in cycles
** per sample.  This acts as a single-frequency discrete Fourier
** transform (DFT) of the input signal.  This transformation, or
** rather a particular efficient algorithm for evaluating it, is known
** as a Goertzel filter (see \cite{jacobsen} or \cite{sahu}).
**
** Immediately, this has the recurrence relation
** \begin{equation}
** g_{n,l} = g_{n,l-1} + x_l\Omega^l - x_{l-n}\Omega^{l-n},
** \label{recrel}
** \end{equation}
** with a straightforward recurrence relation for $\Omega$, given by
** \begin{align*}
** \Omega^{k+1} &= \Omega^k\Omega\\
** \Omega^{k+1}_r &= \Omega^k_r \Omega_r - \Omega^k_i \Omega_i\\
** \Omega^{k+1}_i &= \Omega^k_r \Omega_i + \Omega^k_i \Omega_r,
** \end{align*}
** writing $\Omega_r\equiv\Re\Omega=\cos2\pi\omega$ and $\Omega_i\equiv\Im\Omega=\sin2\pi\omega$,
** and $\Omega^k_r = \Re(\Omega^k)$.  Equation~(\ref{recrel}) is an $n$-point
** `sliding DFT' estimate of the Fourier transform of the input
** signal, at sample~$l$.
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define DTMFX_IMPLEMENTATION 1
#include "goertzel.h"
#include "util.h"

GoertzelFilter* goertzel_init(GoertzelFilter* f,
                              double freq, /* in Hertz (ie cycles per second) */
                              int samp_per_sec, /* samples per second */
                              double filter_length) /* in seconds */
{
    double twopi = 4.0*asin(1);
    double omega;             /* cycles per sample */
    int N;                    /* the filter length, in samples */
    int i;
    
    N = filter_length * samp_per_sec;
    omega = freq/samp_per_sec;
    
    if ((f->xOmegar = (double*)malloc(N*sizeof(double))) == 0) {
        error_exit("Unable to allocate %ld bytes for GoertzelFilter's xr", N*sizeof(double));
    }
    if ((f->xOmegai = (double*)malloc(N*sizeof(double))) == 0) {
        error_exit("Unable to allocate %ld bytes for GoertzelFilter's xi", N*sizeof(double));
    }
    // initialise all of the values of x[] to 0.0
    for (i=0; i<N; i++) {
        f->xOmegar[i] = 0.0;
        f->xOmegai[i] = 0.0;
    }
    f->idx = 0;
    f->N = N;
    f->Omega_nr = 1;
    f->Omega_ni = 0;
    f->Omegar = cos(twopi * omega);
    f->Omegai = sin(twopi * omega);
    f->gr = 0;
    f->gi = 0;

    if (program_verbosity() > 0) {
        fprintf(stderr, "goertzel_init: freq=%g  sample=%gms  N=%d @ %d/s  omega=%g  Omega=(%g,%g)\n",
                freq, filter_length*1000, N, samp_per_sec,
                omega, f->Omegar, f->Omegai);
    }
    
    return f;
}

void goertzel_uninit(GoertzelFilter* f)
{
    free(f->xOmegar);
    free(f->xOmegai);
}

/*
 * Calculate y_l = y_{l-1} exp(2\pi i\omega) + x_l - x_{l-N} exp(2\pi i\omega N)
 */
float goertzel_step(GoertzelFilter* f, float xl)
{
    int idx;
    double gr, gi, Omega_nr, Omega_ni, xOmega_nr, xOmega_ni; /* new values */

    idx = f->idx;
    assert(0 <= idx && idx < f->N);

    Omega_nr = f->Omega_nr * f->Omegar - f->Omega_ni * f->Omegai;
    Omega_ni = f->Omega_nr * f->Omegai + f->Omega_ni * f->Omegar;
    xOmega_nr = xl * Omega_nr;
    xOmega_ni = xl * Omega_ni;

    gr = f->gr + xOmega_nr - f->xOmegar[idx];
    gi = f->gi + xOmega_ni - f->xOmegai[idx];

    f->gr = gr;
    f->gi = gi;
    f->xOmegar[idx] = xOmega_nr;
    f->xOmegai[idx] = xOmega_ni;
    f->Omega_nr = Omega_nr;
    f->Omega_ni = Omega_ni;
    
    idx++;
    if (idx == f->N)
        idx = 0;
    f->idx = idx;
    assert(0 <= idx && idx < f->N);

    /* We end up with f->gr,gi holding Re,Im(g_l) */
    return gr*gr + gi*gi;
}

/*
** \begin{thebibliography}{1}
** \bibitem{jacobsen}
** Jacobsen, Eric and Richard Lyons, \emph{The Sliding DFT}, IEEE
** Signal Processing Magazine, \textbf{20}(2), 74--80 (2003)
** doi:10.1109/MSP.2003.1184347
** \bibitem{sahu}
** Sahu, Anit Kumar and Mrityunjoy Chakraborty, \emph{Fast and
** Accurate Frequency Estimation Using Sliding DFT} (2012). Online:
** http://arxiv.org/abs/1202.4446
** \end{thebibliography}
*/
