#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>
#include <ctype.h>

#include <config.h>

#include "dtmfx.h"
#include "jason/src/jason.h"

static int verbosity = DTMFX_QUIET;

static void Usage(int);
static char* progname;


int main(int argc, char** argv) 
{
    DTMFX_SoundInfo sound_info;
    byte* msg;
    int msglen;
    int output_fd = STDOUT_FILENO;
    char* output_message = NULL;
    int ch;
    int message_is_hex_p = 1;

    progname = argv[0];         /* for reporting below */
    dtmfx_set_program_name(argv[0]);

    // Set up the information about the sound to be produced.
    // These values act as the defaults for the options below.
    sound_info.bits_per_sample = 12;
    sound_info.bytes_per_sample = (sound_info.bits_per_sample + 7)/8;
    sound_info.sample_rate = 11025; /* samples per second */
    sound_info.nchannels = 1;
    sound_info.duration = 0.10;     /* seconds */
    sound_info.gap_duration = 0.05; /* seconds */
    sound_info.start_padding = sound_info.end_padding = 0;
    sound_info.volume = 0.5;

    while ((ch = getopt(argc, argv, "b:hf:gl:o:p:s:v:V?")) != -1) {
        switch (ch) {
          case 'l':
            /*
             * length of beep or length+gap
             * If either figure is <= 1.0, it is in seconds;
             * or if > 1.0, it is in ms
             */
            {
                const char* value = optarg;
                char* endptr;
                float n;
                n = strtof(value, &endptr);
                if (n <= 0.0) {
                    Usage(1);
                }
                if (n > 1.0) {
                    n /= 1000.0; /* convert ms to sec */
                }
                sound_info.duration = n;

                if (*endptr == '\0') {
                    // nothing to do
                } else if (*endptr == '+') {
                    n = strtof(endptr, NULL);
                    if (n <= 0.0) {
                        Usage(1);
                    }
                    if (n > 1.0) {
                        n /= 1000.0;
                    }
                    sound_info.gap_duration = n;
                }
            }
            break;

          case 'p':
            /*
             * Padding: the amount of silence to add before and after the tones
             */
            {
                float n = strtof(optarg, NULL);
                if (n <= 0) {
                    Usage(1);
                }
                if (n > 1.0) {
                    n /= 1000.0; /* convert ms to seconds */
                }
                sound_info.start_padding = sound_info.end_padding = n;
                break;
            }    
                  
          case 'v':
            /*
             * Volume: if <=1, this is a fraction of full volume;
             * if >1, it's a percentage
             */
            {
                const char* value = optarg;
                float n;
                n = strtof(value, NULL);
                if (n <= 0.0) {
                    Usage(1);
                }
                if (n > 1.0) {
                    n /= 100.0; /* convert %age to fraction */
                }
                if (n > 1.0) {  /* ...still */
                    error_exit("Can't have volume greater than 100%%: %f", n*100);
                }
                sound_info.volume = n;
            }
            break;

          case 's':
            /* Sample rate */
            {
                long sample_rate = strtol(optarg, NULL, 10);
                if (sample_rate < 0) {
                    error_exit("Can't have negative sample rate: %d", sample_rate);
                } else if (sample_rate == 0) {
                    Usage(1);
                }
                sound_info.sample_rate = (int)sample_rate;
            }
            break;

          case 'b':
            /* bits per sample */
            if (*optarg == 'f') {
                // floating-point output
                sound_info.bits_per_sample = -32;
                sound_info.bytes_per_sample = 4;
            } else {
                long bits_per_sample = strtol(optarg, NULL, 10);
                if (bits_per_sample <= 1) {
                    error_exit("Minimum bits-per-sample is 2");
                } else if (bits_per_sample <= 32) {
                    sound_info.bits_per_sample = bits_per_sample;
                    sound_info.bytes_per_sample = (sound_info.bits_per_sample + 7)/8;
                } else {
                    error_exit("Maximum bits-per-sample is 32");
                }
            }
            break;

          case 'f':
            /* 'from' format: "hex" or "json" */
            if (strcmp(optarg, "hex") == 0) {
                message_is_hex_p = 1;
            } else if (strcmp(optarg, "json") == 0) {
                message_is_hex_p = 0;
            } else {
                error_exit("Unrecognised -f format %s (may be 'hex' or 'json')",
                           optarg);
            }
            break;

          case 'o':
            /* Output file name */
            {
                int new_fd = open(optarg, O_WRONLY|O_CREAT, 0666);
                if (new_fd >= 0) {
                    output_fd = new_fd;
                } else {
                    error_exit("Can't open file %s to write", optarg);
                }
            }
            break;

          case 'g':
            if (verbosity < DTMFX_DEBUG) {
                verbosity++;
            }
            dtmfx_set_program_verbosity(verbosity);
            break;

          case 'V':
            {
                const char* snapshotid = SNAPSHOTID;
                if (snapshotid[0] == '\0') {
                    // release version
                    printf("This is hex2dtmf.\nFrom %s v%s, released %s\nSee <https://bitbucket.org/nxg/dtmfx>\n",
                           PACKAGE_NAME, PACKAGE_VERSION, RELEASEDATE);
                } else {
                    printf("This is hex2dtmf.\nFrom %s v%s, snapshot %s\n",
                           PACKAGE_NAME, PACKAGE_VERSION, snapshotid);
                }
                exit(0);
            }
            break;

          case 'h':
          case '?':
            Usage(0);
          default:
            Usage(1);
        }
    }
    argc -= optind;
    argv += optind;

    switch (argc) {
      case 0:
        {
            size_t buflen = 64;
            size_t idx = 0;
            char* buf = NULL;
            int got_it = 0;

            while (! got_it) {
                int ch = getchar();
                if (ch == EOF) {
                    assert(idx <= buflen);
                    buf[idx] = '\0';
                    got_it = 1;
                } else if (isxdigit(ch) || ch=='#' || ch=='*') {
                    if (buf == NULL) {
                        if ((buf = (char*)malloc(buflen+1)) == NULL) {
                            error_exit("Can't allocate %ld for input buffer!", buflen);
                        }
                        idx = 0;
                    } else if (idx == buflen) {
                        buflen *= 2;
                        if ((buf = (char*)realloc((void*)buf, buflen+1)) == NULL) {
                            error_exit("Can't reallocate %ld for input buffer!", buflen);
                        }
                    }
                    assert(idx <= buflen);
                    buf[idx++] = ch;
                } else if (isspace(ch)) {
                    // do nothing -- ignore this character
                } else {
                    error_exit("Unexpected character in message: %c", ch);
                }
            }
            output_message = buf;
        }
          break;

      case 1:
        output_message = argv[0];
        break;
        
      default:
        Usage(1);
        break;
    }

    assert(0 < sound_info.duration);
    assert(0 < sound_info.gap_duration);
    assert(0 < sound_info.volume && sound_info.volume <= 1.0);
    assert(0 < sound_info.sample_rate);
    assert((1 < sound_info.bits_per_sample && sound_info.bits_per_sample <= 32)
           || sound_info.bits_per_sample == -32);
    assert(1 <= sound_info.bytes_per_sample && sound_info.bytes_per_sample <= 4);
    assert(output_message != NULL);

    if (verbosity > 0) {
        fprintf(stderr, "dur=%.0f+%.0fms  vol=%.2f  sample-rate=%d  msg=\"%s\"\n",
                sound_info.duration*1000, sound_info.gap_duration*1000,
                sound_info.volume, sound_info.sample_rate,
                output_message);
    }

    if (message_is_hex_p) {
        msglen = strlen(output_message);
        if ((msg = (byte*)malloc(msglen)) == NULL) {
            error_exit("Can't allocate %d bytes of memory!", msglen);
        }
        for (ch=0; ch<msglen; ch++) {
            switch (output_message[ch]) {
              case '0': msg[ch] = DTMF_0; break;
              case '1': msg[ch] = DTMF_1; break;
              case '2': msg[ch] = DTMF_2; break;
              case '3': msg[ch] = DTMF_3; break;
              case '4': msg[ch] = DTMF_4; break;
              case '5': msg[ch] = DTMF_5; break;
              case '6': msg[ch] = DTMF_6; break;
              case '7': msg[ch] = DTMF_7; break;
              case '8': msg[ch] = DTMF_8; break;
              case '9': msg[ch] = DTMF_9; break;
              case 'a': case 'A':
                msg[ch] = DTMF_A; break;
              case 'b': case 'B':
                msg[ch] = DTMF_B; break;
              case 'c': case 'C':
                msg[ch] = DTMF_C; break;
              case 'd': case 'D':
                msg[ch] = DTMF_D; break;
              case 'e': case 'E': case '*':
                msg[ch] = DTMF_E; break;
              case 'f': case 'F': case '#':
                msg[ch] = DTMF_F; break;
              default:
                Usage(1);
            }
        }
    } else {
        // message is JSON
        JsonObject o = parse_json_string(output_message);
        byte* b;
        size_t blen;

        if (o == NULL) {
            error_exit("Can't parse JSON");
        }
        b = get_der_encoding(o, &blen);
        if (b == NULL) {
            error_exit("Can't convert JSON message to DER");
        }
        msglen = 2*blen;
        
        // translate this array of bytes into a 'message' comprising a
        // sequence of nibbles, one hex digit each.  Add them
        // most-significant nibble first.
        if ((msg = (byte*)malloc(msglen)) == NULL) {
            error_exit("Can't allocate %d bytes of memory!", msglen);
        } else {
            byte nibbles[] = { DTMF_0, DTMF_1, DTMF_2, DTMF_3,
                               DTMF_4, DTMF_5, DTMF_6, DTMF_7,
                               DTMF_8, DTMF_9, DTMF_A, DTMF_B,
                               DTMF_C, DTMF_D, DTMF_E, DTMF_F, };
            int i;
            for (i=0; i<blen; i++) {
                msg[2*i] = nibbles[(b[i]&0xf0) >> 4];
                msg[2*i+1] = nibbles[b[i]&0x0f];
            }
        }

        free_json_object(o);
    }
    
    if (verbosity > 0) {
        fprintf(stderr, "parsed message: ");
        for (ch=0; ch<msglen; ch++)
            fprintf(stderr, "%d ", msg[ch]);
        fprintf(stderr, "\n");
    }

    dtmfx_write_wav(output_fd, &sound_info, msg, msglen);

    if (output_fd != STDOUT_FILENO) {
        close(output_fd);
    }

    return 0;
}

static void Usage(int exit_status)
{
    fprintf(stderr, "Usage: %s [-hgV] [-f msgfmt] [-l lenspec] [-p lenspec] [-v volspec] [-b bits] [-o filename] [msg]\n",
            progname);
    exit(exit_status);
}
