/*
 * The WAVE audio standard is described in (for example)
 * http://www-mmsp.ece.mcgill.ca/Documents/AudioFormats/WAVE/WAVE.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>

#define DTMFX_IMPLEMENTATION 1
#include "wav_common.h"
#include "read_wav.h"
#include "util.h"
#include "decoder.h"

#include "config.h"


/* number of samples in a 'frame' */
#define FRAME_SIZE 1024
static int handle_wav_file_float(cbyte *buf, cbyte* endbuf, FmtCkPCMHeader* fmt, void* analyser_state)
{
    int nchannels = fmt->channels;
    int samples_per_sec = fmt->samples_per_sec;
    Sample* p;
    Sample* endp;
    
    assert(sizeof(Sample) == sizeof(float));

    if (nchannels != 1) {
        error_exit("Can only handle one channel right now, not %d", nchannels);
    }

    endp = (Sample*)endbuf;
    
    for (p=(Sample*)buf; p < endp-FRAME_SIZE; p+=FRAME_SIZE) {
        process_frame(analyser_state,
                      p,
                      samples_per_sec,
                      FRAME_SIZE);
    }

    // p is now close to the end of the buffer
    assert(endp-p < FRAME_SIZE);
    process_frame(analyser_state,
                  p,
                  samples_per_sec,
                  endp-p);

    return 0;
}
#undef FRAME_SIZE

#define DTMF_SOUND_WIDTH 1
#include "read_wav.inc"
#undef DTMF_SOUND_WIDTH
#define DTMF_SOUND_WIDTH 2
#include "read_wav.inc"
#undef DTMF_SOUND_WIDTH
#define DTMF_SOUND_WIDTH 3
#include "read_wav.inc"
#undef DTMF_SOUND_WIDTH
#define DTMF_SOUND_WIDTH 4
#include "read_wav.inc"
#undef DTMF_SOUND_WIDTH

static int handle_wav_file(cbyte *p, cbyte* endp, FmtCkPCMHeader* fmt, void* analyser_state)
{
    int bytes_per_sample = 0;
    int retval = 1;             /* error value */

    if (fmt->format_tag == 1) {
        bytes_per_sample = (fmt->bits_per_sample + 7)/8;
    } else if (fmt->format_tag == 3) {
        // floating point samples
        bytes_per_sample = -4;
    } else {
        error_exit("Unexpected format tag: %d", fmt->format_tag);
    }
    assert(bytes_per_sample != 0);

    switch (bytes_per_sample) {
      case 1:
        retval = handle_wav_file_1(p, endp, fmt, analyser_state);
        break;

      case 2:
        retval = handle_wav_file_2(p, endp, fmt, analyser_state);
        break;

      case 3:
        retval = handle_wav_file_3(p, endp, fmt, analyser_state);
        break;

      case 4:
        retval = handle_wav_file_4(p, endp, fmt, analyser_state);
        break;

      case -4:
        retval = handle_wav_file_float(p, endp, fmt, analyser_state);
        break;

      default:
        error_exit("Unexpected number of bytes-per-sample: %d", bytes_per_sample);
        break;
    }

    return retval;
}

int read_audio_wav(const int fd, const char* filename, void* analyser_state)
{
    cbyte *wav_file = NULL;
    cbyte *wav_p;
    cbyte *wav_endp;
    
    size_t filesize;
    ChunkHeader *riff_chunk;
    char* wave_id;
    FmtCkPCMHeader *fmt_chunk;
    ChunkHeader *data_chunk;
    int chatter_p = (program_verbosity() > 0);

    if (fd >= 0) {
        error_exit("Can't read WAV from FD yet");
    } else if (filename != NULL) {
        wav_file = map_file(filename, &filesize);
    } else {
        error_exit("Bad call to read_audio_wav with no fd or filename");
    }
    assert(wav_file != NULL);

    if (strncmp((char*)wav_file, "RIFF", 4) != 0) {
        if (program_verbosity() > 0) {
            fprintf(stderr, "File %s is not a WAV file\n", filename);
        }
        unmap_file(wav_file, filesize);
        return 1;
    }

    wav_p = wav_file;
    wav_endp = &wav_file[filesize];

    riff_chunk = (ChunkHeader*)wav_p;
    if (chatter_p) {
        fprintf(stderr, "WAV header: id=%.4s  size=%d\n", riff_chunk->ckid, riff_chunk->cksize);
    }
    wav_p += sizeof(ChunkHeader);

    wave_id = (char*)wav_p;
    if (strncmp(wave_id, "WAVE", 4) != 0) {
        error_exit("Unexpected id at top of WAV file (expecting \"WAVE\"): %.4s", wave_id);
    }
    wav_p += 4;

    fmt_chunk = (FmtCkPCMHeader*)wav_p;

    
    if (chatter_p) {
        fprintf(stderr, "WAV format: size=%d  channels=%d  samples/sec=%d  bits/sample=%d  format=%d\n",
                fmt_chunk->cksize, fmt_chunk->channels,
                fmt_chunk->samples_per_sec, fmt_chunk->bits_per_sample, fmt_chunk->format_tag);
    }
    wav_p += sizeof(FmtCkPCMHeader);
    
    data_chunk = (ChunkHeader*)wav_p;
    if (strncmp(data_chunk->ckid, "data", 4) != 0) {
        error_exit("Unexpected tag at top of WAV file (expecting \"data\"): %.4s", data_chunk->ckid);
    }
    wav_p += sizeof(ChunkHeader);

    handle_wav_file(wav_p, wav_endp, fmt_chunk, analyser_state);

    return 0;
}
