#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define DTMFX_IMPLEMENTATION 1
#include "util.h"

/*
 * Produce an error message, and exit with a non-zero status.
 * If the first character of 'fmt' is '!', then abort instead.
 */
void error_exit(const char* fmt, ...)
{
    va_list ap;
    int abort_p = 0;
    const char* actual_fmt = fmt;
    
    va_start(ap, fmt);

    if (*fmt == '!') {
        abort_p = 1;
        actual_fmt = &fmt[1];
    }
    
    fprintf(stderr, "%s: ", program_name());
    vfprintf(stderr, actual_fmt, ap);
    if (abort_p) {
        fprintf(stderr, " [this shouldn't happen]");
    }
    fprintf(stderr, "\n");
    va_end(ap);

    if (abort_p) {
        abort();                /* flushes and closes streams */
    } else {
        exit(1);
    }
}

static const char* _progname;
void dtmfx_set_program_name(const char* s)
{
    _progname = s;
}
const char* program_name(void)
{
    return _progname;
}

static int _verbosity;
void dtmfx_set_program_verbosity(int i)
{
    _verbosity = i;
}
int program_verbosity(void)
{
    return _verbosity;
}

const unsigned char* map_file(const char* filename, size_t* file_size)
{
    struct stat S;
    int fileno;
    void* mapped_file;

    if ((fileno = open(filename, O_RDONLY)) < 0) {
        error_exit("Can't open file %s to read", filename);
    }
    if (fstat(fileno, &S) == -1 || S.st_size == 0) {
        error_exit("Can't stat open file %s", filename);
    }
    mapped_file = mmap(0, S.st_size, PROT_READ, MAP_FILE|MAP_PRIVATE, fileno, 0);
    if (mapped_file == MAP_FAILED) {
        close(fileno);
        error_exit("Unable to map open file %s", filename);
    }

    if (file_size != NULL) {
        *file_size = S.st_size;
    }

    return (const unsigned char*)mapped_file;
}

void unmap_file(const unsigned char* mapped_file, size_t file_size)
{
    munmap((void*)mapped_file, file_size);
}


