/*
 * The WAVE audio standard is described in (for example)
 * http://www-mmsp.ece.mcgill.ca/Documents/AudioFormats/WAVE/WAVE.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>

#define DTMFX_IMPLEMENTATION 1
#include "wav_common.h"
#include "write_wav.h"
#include "util.h"

#include "config.h"

/* Deliberately ignore the return value of write(), since some
   compilers (some gcc versions) complain about doing so, and
   (void)write() doesn't stop the message. */
#define WRITE(args...) \
    { ssize_t rval = write(args);                                       \
        if (rval < 0) {                                                 \
        fprintf(stderr,                                                 \
                "write produced unexpected error on line __LINE__ (ignored)\n"); \
        }                                                               \
    }

#ifndef BUFSIZ
/* setbuf(3) indicates that BUFSIZ (when declared) is optimal */
#define BUFSIZ 1024
#endif

// Define functions write_audio_wav{1,2,4,iee754}
#define DTMF_SOUND_WIDTH 1
#include "write_wav.inc"
#undef DTMF_SOUND_WIDTH

#define DTMF_SOUND_WIDTH 2
#include "write_wav.inc"
#undef DTMF_SOUND_WIDTH

#define DTMF_SOUND_WIDTH 4
#include "write_wav.inc"
#undef DTMF_SOUND_WIDTH

#define DTMF_SOUND_WIDTH -32
#include "write_wav.inc"
#undef DTMF_SOUND_WIDTH

int write_audio_wav(int output_fd, DTMFX_SoundInfo* sound_info, byte* msg, int nmsg)
{
    int rval;
    
    assert(sound_info->bytes_per_sample > 0);
    
    if (sound_info->bits_per_sample == -32) {
        rval = write_audio_wav_ieee754(output_fd, sound_info, msg, nmsg);
    } else if (sound_info->bits_per_sample < 0) {
        error_exit("Impossible audio bit-width: %d", sound_info->bits_per_sample);
    } else if (sound_info->bytes_per_sample == 1) {
        rval = write_audio_wav1(output_fd, sound_info, msg, nmsg);
    } else if (sound_info->bytes_per_sample == 2) {
        rval = write_audio_wav2(output_fd, sound_info, msg, nmsg);
    } else if (sound_info->bytes_per_sample <= 4) {
        rval = write_audio_wav4(output_fd, sound_info, msg, nmsg);
    } else {
        error_exit("Impossible value of bytes_per_sample: %d", sound_info->bytes_per_sample);
    }
    return rval;
}

int write_header_wav(int output_fd, DTMFX_SoundInfo* sound_info, int nmsg)
{
    ChunkHeader riff_chunk;
    char WAVE_ID[4] = "WAVE";
    FmtCkPCMHeader fmt_chunk;
    ChunkHeader data_chunk;

    int nchannels = sound_info->nchannels;
    int sample_rate = sound_info->sample_rate;
    int bits_per_sample = sound_info->bits_per_sample;
    int bytes_per_sample = sound_info->bytes_per_sample;
    float duration = nmsg * sound_info->duration
        + (nmsg-1)*sound_info->gap_duration
        + sound_info->start_padding
        + sound_info->end_padding;

    // Sanity checks -- some of these would be better as compile-time
    // assertions, but that's tricky in C with sizeof 
    assert(bits_per_sample == -32 || (bits_per_sample > 2 && bits_per_sample <= 32));
    assert(sizeof(struct FmtCkPCMHeader_s) == 24);
    
    strncpy(fmt_chunk.ckid, "fmt ", 4);
    fmt_chunk.cksize = sizeof(FmtCkPCMHeader)-8;
    fmt_chunk.channels = nchannels;
    fmt_chunk.samples_per_sec = sample_rate;
    fmt_chunk.avg_bytes_per_sec = nchannels * sample_rate * bytes_per_sample;
    fmt_chunk.block_align = nchannels * bytes_per_sample;
    if (bits_per_sample == -32) {
        fmt_chunk.bits_per_sample = 32;
        fmt_chunk.format_tag = 3;   /* Pulse Code Modulation, floating point */
    } else {
        fmt_chunk.bits_per_sample = bits_per_sample;
        fmt_chunk.format_tag = 1;   /* Microsoft Pulse Code Modulation */
    }

    strncpy(data_chunk.ckid, "data", 4);
    data_chunk.cksize = lroundf(nchannels*duration*sample_rate*bytes_per_sample);

    if (program_verbosity() > 0) {
        fprintf(stderr, "duration=%.3fs  data-size=%d\n", duration, data_chunk.cksize);
    }

    strncpy(riff_chunk.ckid, "RIFF", 4);
    riff_chunk.cksize = fmt_chunk.cksize + 8
        + data_chunk.cksize + 8
        + 4;

    WRITE(output_fd, &riff_chunk, sizeof(riff_chunk));
    WRITE(output_fd, &WAVE_ID, 4);
    WRITE(output_fd, &fmt_chunk,  sizeof(fmt_chunk));
    WRITE(output_fd, &data_chunk, sizeof(data_chunk));

    return 0;
}

/**
 * Write a 'message' of bytes to the given output FD.  The 'message'
 * is an array, MSG, which is NMSG bytes long; each byte in the
 * array is one of the constants DTMF_0 to DTMF_F.
 *
 * @return 0 on success
 */
int dtmfx_write_wav(int output_fd, DTMFX_SoundInfo* sound_info, byte* msg, int nmsg)
{
    write_header_wav(output_fd, sound_info, nmsg);
    return write_audio_wav(output_fd, sound_info, msg, nmsg);
}
