/*
 * This file contains implementations for WAV-encoding for multiple
 * bit-widths, since the mechanics of the encoding for these cases is
 * almost, but not quite, the same.
 *
 * Highlights of the differences:
 *
 *   * 1-byte sounds is unsigned, with a 0x80 offset
 *
 *   * 3-byte sound is handled here as 4-byte sound (there is no
 *     int24_t type!), but with bit-twiddling as appropriate)
 *
 *   * the floating-point representation is flagged with a 'bit-width'
 *     of -32
 *
 * The including file must define DTMF_SOUND_WIDTH=1, 2, or 4
 */
#if DTMF_SOUND_WIDTH == 1
#define MAKE_WAVEFORM    make_waveform1
#define WRITE_AUDIO_WAV  write_audio_wav1
#define DESTTYPE         byte
#define BITWIDTH         8
#define BYTES_PER_SAMPLE 1
#elif DTMF_SOUND_WIDTH == 2
#define MAKE_WAVEFORM    make_waveform2
#define WRITE_AUDIO_WAV  write_audio_wav2
#define DESTTYPE         int16_t
#define BITWIDTH         16
#define BYTES_PER_SAMPLE 2
#elif DTMF_SOUND_WIDTH == 4
#define MAKE_WAVEFORM    make_waveform4
#define WRITE_AUDIO_WAV  write_audio_wav4
#define DESTTYPE         int32_t
#define BITWIDTH         32
#define BYTES_PER_SAMPLE 4
#elif DTMF_SOUND_WIDTH == -32
/* IEEE-754 float */
#define MAKE_WAVEFORM    make_waveform_ieee754
#define WRITE_AUDIO_WAV  write_audio_wav_ieee754
#define DESTTYPE         float
#define BITWIDTH         32
#define BYTES_PER_SAMPLE 4
#else
#error "Impossible value of DTMF_SOUND_WIDTH"
#endif

static DESTTYPE* MAKE_WAVEFORM(DTMFX_SoundInfo* si,
                               int freq,
                               float max_ampl_fraction,
                               int* nsamples);
static int WRITE_AUDIO_WAV(int output_fd,
                           DTMFX_SoundInfo* sound_info,
                           byte* msg,
                           int nmsg);

// In order to avoid some rounding errors, we generate multiple cycles
// of the single-tone sample waveforms.
#define SAMP_DUPL 16.0

static DESTTYPE* MAKE_WAVEFORM(DTMFX_SoundInfo* si,
                               int freq,
                               float max_ampl_fraction,
                               int* nsamples)
{
    int nsx = lround((SAMP_DUPL * si->sample_rate)/freq);
    float ns = (1.0 * si->sample_rate)/freq;
    int bps = si->bits_per_sample;
    float twopi;
    float max_ampl;
#if DTMF_SOUND_WIDTH == 1
    float offset = 64;
#endif
    int i;
    DESTTYPE* buf;
    int unused_bits = BITWIDTH - bps;

#if DTMF_SOUND_WIDTH == 4
    assert(si->bytes_per_sample == 3 || si->bytes_per_sample == 4);
#else
    assert(si->bytes_per_sample == BYTES_PER_SAMPLE);
#endif
    assert(unused_bits >= 0);

    twopi = 4.0 * asin(1);
    // halve the max amplitude so that we get the full amplitude when
    // we add two waveforms together
#if DTMF_SOUND_WIDTH == -32
    // 'full scale' is 1.0
    max_ampl = max_ampl_fraction/2.0;
#else
    max_ampl = (1<<(bps - 1)) * max_ampl_fraction/2.0;
#endif
    *nsamples = nsx;

    //fprintf(stderr, "max_ampl=%f(%.1f), bps=%d, unused=%d, ns=%d/%f\n", max_ampl, max_ampl_fraction, bps, unused_bits, nsx, ns);
    
    if ((buf = (DESTTYPE*)malloc(nsx * BYTES_PER_SAMPLE)) == NULL) {
        error_exit("Can't allocate %d bytes", nsx * BYTES_PER_SAMPLE);
    }
    for (i=0; i<nsx; i++) {
#if DTMF_SOUND_WIDTH == -32
        buf[i] = max_ampl * sinf((i*twopi)/ns);
#else
        buf[i] = lround(max_ampl * sinf((i*twopi)/ns)
#if DTMF_SOUND_WIDTH == 1
                + offset
#endif
                ) << unused_bits;
#endif
    }
    
    return buf;
}


int WRITE_AUDIO_WAV(int output_fd, DTMFX_SoundInfo* sound_info, byte* msg, int nmsg)
{
    int i;
    int msg_idx;
    const int bytes_per_sample = sound_info->bytes_per_sample;

    int low_nsamples[4];
    DESTTYPE* low_samples[4];
    int high_nsamples[4];
    DESTTYPE* high_samples[4];
    /* what might be the optimal buffer size? */
    DESTTYPE buffer[BUFSIZ];
    byte zerobuf[BUFSIZ];

#if DTMF_SOUND_WIDTH == 4
    assert(bytes_per_sample == 3 || bytes_per_sample == 4);
#else
    assert(bytes_per_sample == BYTES_PER_SAMPLE);
#endif

#if DTMF_SOUND_WIDTH == 1
    memset(zerobuf, 0x80, BUFSIZ);
#else
    memset(zerobuf, 0, BUFSIZ);
#endif

    /* Now write out the samples */
    for (i=0; i<4; i++) {
        low_samples[i] = MAKE_WAVEFORM(sound_info,
                                       dtmf_low_freq(i),
                                       sound_info->volume,
                                       &low_nsamples[i]);
    }
    for (i=0; i<4; i++) {
        high_samples[i] = MAKE_WAVEFORM(sound_info,
                                        dtmf_high_freq(i),
                                        sound_info->volume,
                                        &high_nsamples[i]);
    }

    for (msg_idx=0; msg_idx<nmsg; msg_idx++) {
        int lowi;
        int highi;
#if DTMF_SOUND_WIDTH == 4
        byte *bp;
        const int buf3max = 3 * (BUFSIZ/3);
#endif
        int bufidx;
        int samp;
        const int nsamp = lroundf(sound_info->duration * sound_info->sample_rate);

        const int fl = msg[msg_idx] & 0x3;      /* low frequency index */
        const int fh = (msg[msg_idx] & 0xc)>>2; /* high frequency index */

        const DESTTYPE* l = low_samples[fl];
        const int nl = low_nsamples[fl];
        const DESTTYPE* h = high_samples[fh];
        const int nh = high_nsamples[fh];

        /* fprintf(stderr, "p=%x  fl=%d (%d samples)  fh=%d (%d); nsamp=%d\n", */
        /*        msg[msg_idx], fl, nl, fh, nh, nsamp); */
        
        if (msg_idx == 0) {
            int nbytes = lroundf(sound_info->start_padding * sound_info->sample_rate) * bytes_per_sample;
            while (nbytes > 0) {
                WRITE(output_fd, zerobuf, (nbytes < BUFSIZ ? nbytes : BUFSIZ));
                nbytes -= BUFSIZ;
            }
        } else if (msg_idx > 0) {
            // not first time -- add gap of silence
            int nbytes = lroundf(sound_info->gap_duration * sound_info->sample_rate) * bytes_per_sample;
            while (nbytes > 0) {
                WRITE(output_fd, zerobuf, (nbytes < BUFSIZ ? nbytes : BUFSIZ));
                nbytes -= BUFSIZ;
            }
        }

        lowi = highi = 0;
#if DTMF_SOUND_WIDTH == 4
        bp = (byte*)buffer;
#endif
        bufidx = 0;
        for (samp=0; samp<nsamp; samp++) {
#if DTMF_SOUND_WIDTH == 4
            if (bytes_per_sample == 3) {
                // in (only) this case, we have to do some
                // bit-twiddling to get the int32_t value we calculate
                // into the next three bytes, little-endian
                int32_t i32 = l[lowi] + h[highi];
                *bp++ = (i32 & 0x0000ff00) >> 8;
                *bp++ = (i32 & 0x00ff0000) >> 16;
                *bp++ = (i32 & 0xff000000) >> 24;
                if (bp - (byte*)buffer > buf3max) {
                    WRITE(output_fd, buffer, bp-(byte*)buffer);
                    bp = (byte*)buffer;
                }   
            } else {
                // this part is the same as the 'normal' cases, as below
                buffer[bufidx] = l[lowi] + h[highi];
                if (++bufidx == BUFSIZ) {
                    WRITE(output_fd, buffer, sizeof(buffer));
                    bufidx = 0;
                }
            }
#else
            buffer[bufidx] = l[lowi] + h[highi];
            if (++bufidx == BUFSIZ) {
                WRITE(output_fd, buffer, sizeof(buffer));
                bufidx = 0;
            }
#endif
            if (++lowi == nl)
                lowi = 0;
            if (++highi == nh)
                highi = 0;
        }

        // write out any remaining contents of the buffer
#if DTMF_SOUND_WIDTH == 4
        if (bytes_per_sample == 3) {
            if (bp > (byte*)buffer) {
                WRITE(output_fd, buffer, bp-(byte*)buffer);
            }
        } else {
            if (bufidx > 0) {
                WRITE(output_fd, buffer, bufidx*sizeof(buffer[0]));
            }
        }
#else
        if (bufidx > 0) {
            WRITE(output_fd, buffer, bufidx*sizeof(buffer[0]));
        }
#endif
    }

    if (sound_info->end_padding > 0) {
        int nbytes = lroundf(sound_info->end_padding * sound_info->sample_rate) * bytes_per_sample;
        while (nbytes > 0) {
            WRITE(output_fd, zerobuf, (nbytes < BUFSIZ ? nbytes : BUFSIZ));
            nbytes -= BUFSIZ;
        }
    }
    return 0;
}

#undef MAKE_WAVEFORM
#undef WRITE_AUDIO_WAV
#undef DESTTYPE
#undef BITWIDTH
#undef BYTES_PER_SAMPLE
