/*
 * This file is part of the Jason library.
 * See <https://bitbucket.org/nxg/jason>
 *
 * Copyright 2013, Norman Gray <http://nxg.me.uk>
 *
 * It is distributed under the terms of the GNU General Public
 * Licence.  See the file README.me for details.
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>
#include <setjmp.h>
#include <math.h>

#include <config.h>
#define JASON_IMPLEMENTATION 1
#include "jason.h"
#include "util.h"

#define DBG_INTEGER 1
#define DBG_REAL 2
#define DBG_SEQUENCES 4
/* #define DEBUG DBG_INTEGER|DBG_REAL|DBG_SEQUENCES */
#define DEBUG 0

static jmp_buf unsupported_jmp;
// The flag in_jmp_code is true when we're in an extent protected by
// unsupported_jmp.  This doesn't have any effect other than letting
// us check that we've nested calls to setjmp/longjmp appropriately.
//
// Of course, this setup makes this whole thing completely non-thread-safe.
static int in_jmp_code = 0;
#define UNSUPPORTED_MSG_LENGTH 128
static char unsupported_msg[UNSUPPORTED_MSG_LENGTH];
static void unsupported_operation(const char *msg, ...)
{
    va_list ap;

    assert(in_jmp_code);
    va_start(ap, msg);
    vsnprintf(unsupported_msg, UNSUPPORTED_MSG_LENGTH, msg, ap);
    va_end(ap);
    longjmp(unsupported_jmp, 1);
}

// If get_der_encoding fails because of unsupported behaviour,
// then it returns NULL and sets a message which can be retrieved
// here.  The message will be NULL unless this is the cause of the failure.
const char* unsupported_message() 
{
    if (unsupported_msg[0] == '\0') {
        return NULL;
    } else {
        return unsupported_msg;
    }
}

JsonObject initialise_json_object(JsonObject r, JsonType type)
{
    r->type = type;
    r->der_encoding = NULL;
    r->der_encoding_length = 0;
    return r;
}

static void free_json_object_internal(JsonObject o, int free_contents)
{
    if (o == NULL)
        return;

    if (free_contents) {
        switch (o->type) {
          case JSON_STRING:
            free((void*)o->x.s);
            break;
          case JSON_KVPAIR:
            free((void*)o->x.kv.k);
            free_json_object_internal(o->x.kv.v, 1);
            break;
          case JSON_OBJECT:
          case JSON_ARRAY:
            {
                int i;
                JsonObject* p = o->x.a.p;
                size_t np = o->x.a.len;
                
                for (i=0; i<np; i++) {
                    free_json_object_internal(p[i], 1);
                }
            }
            break;
          default:
            // nothing to do
            break;
        }
    }
    
    free((void*)o);
}

/**
 * Free the object.
 * Any allocated memory within the object is freed also.
 */
void free_json_object(JsonObject o)
{
    free_json_object_internal(o, 1);
}

// JSON object to string
//
// We make some efforts to support round-tripping, but this is
// slightly complicated for reals, enough that it's not really worth
// bothing about getting it perfect.
static void print_json_array(JsonObject arr, StringBuilder sb);
static void print_json_oneobject(JsonObject obj, StringBuilder sb);

void print_json_object_to_buffer(JsonObject obj, StringBuilder sb)
{
#define BUFLEN 64          /* can any integer or float be longer than this? */
    char printbuf[BUFLEN];
    
    switch (obj->type) {
      case JSON_STRING:
        {
            const char* sp = obj->x.s;
            const char* ep;           /* character to be escaped */
            append_to_buffer_c(sb, '\"');
            while (sp != NULL) {
                ep = strpbrk(sp, "\"\\/\b\f\n\r\t");
                if (ep == NULL) {
                    append_to_buffer_s(sb, sp);
                    sp = NULL;
                } else {
                    char esc;
                    switch (*ep) {
                      case '\"': esc = '\"'; break;
                      case '\\': esc = '\\'; break;
                      case '/':  esc = '/'; break;
                      case '\b': esc = 'b'; break;
                      case '\f': esc = 'f'; break;
                      case '\n': esc = 'n'; break;
                      case '\r': esc = 'r'; break;
                      case '\t': esc = 't'; break;
                      default:
                        assert(0);
                    }
                    append_to_buffer_n(sb, (const byte*)sp, ep-sp);
                    append_to_buffer_c(sb, '\\');
                    append_to_buffer_c(sb, esc);
                    sp = ep+1;
                }
            }
            append_to_buffer_c(sb, '\"');
        }
        break;
      case JSON_INTEGER:
        snprintf(printbuf, BUFLEN, "%" JSONFMT, obj->x.i);
        append_to_buffer_s(sb, printbuf);
        break;
      case JSON_REAL:
        if (nearbyint(obj->x.f) == obj->x.f) {
            // this is an integer value, so force there to be a
            // trailing ".0" so that this will be reparsed as a real
            // rather than an integer.
            snprintf(printbuf, BUFLEN, "%.0f.0", obj->x.f);
        } else {
            snprintf(printbuf, BUFLEN, "%g", obj->x.f);
        }
        append_to_buffer_s(sb, printbuf);
        break;
      case JSON_ARRAY:
        print_json_array(obj, sb);
        break;
      case JSON_OBJECT:
        print_json_oneobject(obj, sb);
        break;
      case JSON_BOOLEAN:
        append_to_buffer_s(sb, (obj->x.b ? "true" : "false"));
        break;
      case JSON_NULL:
        append_to_buffer_s(sb, "null");
        break;
      default:
        error_exit("Unexpected object of type %d\n", obj->type);
    }
#undef BUFLEN
}

static void print_json_array(JsonObject arr, StringBuilder sb)
{
    JsonObject* p = arr->x.a.p;
    size_t np = arr->x.a.len;
    int i;
    int first_time = 1;
    
    append_to_buffer_s(sb, "[ ");
    for (i=0; i<np; i++) {
        if (first_time)
            first_time = 0;
        else
            append_to_buffer_s(sb, ", ");
        
        print_json_object_to_buffer(p[i], sb);
    }
    append_to_buffer_s(sb, " ]");
}

static void print_json_oneobject(JsonObject obj, StringBuilder sb)
{
    JsonObject* p = obj->x.a.p;
    size_t np = obj->x.a.len;
    int i;
    int first_time = 1;

    append_to_buffer_s(sb, "{ ");
    for (i=0; i<np; i++) {
        if (first_time)
            first_time = 0;
        else
            append_to_buffer_s(sb, ", ");

        append_to_buffer_c(sb, '\"');
        append_to_buffer_s(sb, p[i]->x.kv.k);
        append_to_buffer_s(sb, "\": ");
        print_json_object_to_buffer(p[i]->x.kv.v, sb);
    }
    append_to_buffer_s(sb, " }");
}

/**
 * Serialise the object to a string.
 *
 * The string must be subsequently freed by the caller.
 */
const char* print_json_object(JsonObject obj)
{
    const char* rval;

    if (obj == NULL) {
        return NULL;
    }
    
    StringBuilder sb = make_append_buffer();
    print_json_object_to_buffer(obj, sb);
    append_to_buffer_c(sb, '\0');
    rval = (const char*)sb->buf;
    free(sb);                   /* doesn't free sb->buf */
    return rval;
}

// creating JSON objects

static JsonObject new_object(JsonType type)
{
    static JsonObject r;
    if ((r = (JsonObject)malloc(sizeof(struct json_object))) == NULL) {
        error_exit("Unable to allocate memory for JSON object!");
    }
    return initialise_json_object(r, type);
}

// argument 's' should me malloc-ed, and not a constant string
JsonObject make_string(const char* s)
{
    JsonObject r = new_object(JSON_STRING);
    r->x.s = s;
    return r;
}

JsonObject make_integer(jsonint i)
{
    JsonObject r = new_object(JSON_INTEGER);
    r->x.i = i;
    return r;
}

JsonObject make_real(double f)
{
    JsonObject r = new_object(JSON_REAL);
    r->x.f = f;
    return r;
}

JsonObject make_boolean(int b)
{
    JsonObject r = new_object(JSON_BOOLEAN);
    r->x.b = b;
    return r;
}

JsonObject make_null(void)
{
    JsonObject r = new_object(JSON_NULL);
    return r;
}

// Create an array object which contains (only) the given value.  This
// is used for both JSON 'array' objects (in which case the initial
// object can be anything) and for JSON 'object' objects (in which
// case the initial object will be a k-v pair).
//
// The 'initial' object is now 'owned' by this module.
static JsonObject make_initial_array(JsonType type, JsonObject initial)
{
    JsonObject r = new_object(type);
    if (initial == NULL) {
        // create an object which is, and will remain, empty
        r->x.a.p = NULL;
        r->x.a.alloc = 0;
        r->x.a.len = 0;
    } else {
        size_t alloc = 4;
        JsonObject* arr = (JsonObject*)malloc(alloc * sizeof(JsonObject));
        if (arr == NULL) {
            error_exit("Can't allocate space for make_array");
        }
        arr[0] = initial;
        r->x.a.p = arr;
        r->x.a.alloc = alloc;
        r->x.a.len = 1;
    }
    return r;
}

JsonObject make_array(JsonObject obj)
{
    return make_initial_array(JSON_ARRAY, obj);
}

JsonObject make_object(JsonObject obj)
{
    assert(obj == NULL || obj->type == JSON_KVPAIR);
    return make_initial_array(JSON_OBJECT, obj);
}

// argument 'k' should me malloc-ed, and not a constant string
JsonObject make_kv(const char* k, JsonObject o)
{
    JsonObject r = new_object(JSON_KVPAIR);
    r->x.kv.k = k;
    r->x.kv.v = o;
    return r;
}

// the newobj object is now 'owned' by this module
JsonObject append_object(JsonObject o, JsonObject newobj)
{
    assert(o != NULL);
    assert(o->type == JSON_ARRAY
           || (o->type == JSON_OBJECT && newobj->type == JSON_KVPAIR));

    // the object/array must have been initialised as non-empty:
    // if this is false, deem it to be a programming error
    assert(o->x.a.alloc > 0);
    if (o->x.a.len == o->x.a.alloc) {
        size_t alloc = 3 * o->x.a.alloc / 2;
        JsonObject* arr = (JsonObject*)realloc((void*)(o->x.a.p),
                                               alloc*sizeof(JsonObject));
        if (arr == NULL) {
            error_exit("Can't realloc memory for append_object");
        }
        o->x.a.p = arr;
        o->x.a.alloc = alloc;
    }
    assert(o->x.a.len < o->x.a.alloc);
    o->x.a.p[o->x.a.len] = newobj;
    o->x.a.len++;

    return o;
}

// DER-encoding

static byte* get_der_encoding_internal(JsonObject obj, size_t *len);
static byte* get_der_encoding_charp(const char* str, size_t *len);

/*
 * Allocate bytes, calling error_exit on any errors
 */
static byte* allocb(size_t len)
{
    byte* b;
    if ((b = (byte*)malloc(len)) == NULL) {
        error_exit("allocb: unable to allocate %ld\n", len);
    }
    return b;
}

/*
 * Allocate a DER object with the given identifier and content length,
 * and fill in its length octet(s).
 */
static byte* alloc_der_object(byte identifier,
                              size_t content_length,
                              size_t* headlen)
{
    byte* b;
    int i;
    // lenbytes is the number of extra length bytes
    // which are required for the length specification
    size_t lenbytes;
    size_t totallen;

    // this module does not include any high-tag-number encodings
    // (a high-tag-number identifier has bits 5-1 all 1).
    assert((identifier & 0x1f) < 0x1f);

    if (content_length <= 127) {
        lenbytes = 0;
    } else {
        lenbytes = 1;
        while (content_length >= (1<<(8*lenbytes))) {
            lenbytes++;
            // as a sanity check, assume that the string length is at
            // most eight bytes (the size of a long integer /
            // size_t).  If this assertion fails, then we've
            // presumably made some signed/unsigned error somewhere.
            assert(lenbytes <= 8);
        }
    }

    totallen = 2 + lenbytes + content_length;
    
    b = allocb(totallen);
    b[0] = identifier;
    if (lenbytes == 0) {
        b[1] = content_length;
    } else {
        size_t l = content_length;
        
        b[1] = (lenbytes | 0x80);
        for (i=2+lenbytes-1; i>=2; i--) {
            b[i] = (l & 0xff);
            l >>= 8;
        }
    }

    if (headlen != NULL) {
        *headlen = 2 + lenbytes;
    }

    return b;
}

/**
 * Get the DER-encoding of the object.
 * The object must not be freed by the caller.
 */
byte* get_der_encoding(JsonObject obj, size_t *len) 
{
    byte* rval;
    
    if (obj == NULL) {
        return NULL;
    }

    assert(! in_jmp_code);
    
    if (setjmp(unsupported_jmp) != 0) {
        // something has called unsupported_operation()
        rval = NULL;
    } else {
        // normal behaviour
        in_jmp_code = 1;
        unsupported_msg[0] = '\0';

        rval = get_der_encoding_internal(obj, len);
    }

    in_jmp_code = 0;

    return rval;
}

/**
 * Utility method to serialise a byte sequence to a printable string of
 * hex digits.  The returned object must be freed by the caller.
 */
char* bytes_to_string(byte* b, size_t blen)
{
    static char* buf = NULL;
    static size_t buflen = 0;
    int i;

    if (b == NULL) {
        return NULL;
    }
    
    if (buf == NULL) {
        buflen = (2*blen)+1;
        if ((buf = (char*)malloc(buflen)) == NULL) {
            error_exit("bytes_to_string: can't malloc %ld bytes\n", buflen);
        }
    } else if (buflen <= (2*blen)) {
        buflen = 2*blen+1;
        if ((buf = (char*)realloc((void*)buf, buflen)) == NULL) {
            error_exit( "bytes_to_string: can't realloc %ld bytes\n", buflen);
        }
    }
    assert(buflen > 2*blen);
    for (i=0; i<blen; i++) {
        sprintf(&buf[2*i], "%02x", b[i]);
    }
    buf[2*i] = '\0';
    return buf;
}

static byte* get_der_encoding_charp(const char* str, size_t *len)
{
    // 0x16 <len> bytes...
    byte* b;
    size_t slen = strlen(str);
    size_t headlen;
    int i;

    b = alloc_der_object(0x16, slen, &headlen);

    for (i=0; i<slen; i++) {
        b[headlen+i] = (byte)(str[i]);
    }

    if (len != NULL) {
        *len = headlen + slen;
    }
    return b;
}

static void set_der_encoding_string(JsonObject obj)
{
    // ASCII only at the moment
    // 0x16 <len> bytes...

    size_t slen;

    assert(obj->type == JSON_STRING);
    obj->der_encoding = get_der_encoding_charp(obj->x.s, &slen);
    obj->der_encoding_length = slen;
}

static byte* encode_integer(jsonint num, size_t *nbyte)
{
    static byte b[8];
    int ndigs;
    int i;
    jsonint one = 1;
    uintmax_t unum;

#if DEBUG & DBG_INTEGER
    printf("encode_integer: num=%" JSONFMT, num);
#endif
    if (num > 0) {
        for (ndigs=1; (one<<(8*ndigs-1)) - num <= 0; ndigs++) {
            // nothing
        }
        assert(ndigs <= sizeof(jsonint));
        unum = num;
    } else {
        for (ndigs=1; (one<<(8*ndigs-1)) + num < 0; ndigs++) {
            // nothing
        }
        assert(ndigs <= sizeof(jsonint));
        if (ndigs == sizeof(jsonint)) {
            // exploit the machines' architecture being
            // twos-complement, and simply copy the bits
            // (attempting the 'else' calculation would overflow)
            uintmax_t* p = (uintmax_t*)&num;
            unum = *p;
        } else {
            unum = (one<<(8*ndigs)) + num;
        }
        /* now positive, and twos-complement with ndigs digits */
    }
#if DEBUG & DBG_INTEGER
    printf(" --> %" PRIuMAX "/%" PRIxMAX " (%d) -->", unum, unum, ndigs);
#endif
   
    *nbyte = ndigs;
    
    for (i=ndigs-1; i>=0; i--) {
        b[i] = (unum & 0xff);
        unum >>= 8;
    }
#if DEBUG & DBG_INTEGER
    for (i=0; i<ndigs; i++) {
        printf(" %02x", b[i]);
    }
    printf("\n");
#endif

    return b;
}

static void set_der_encoding_integer(JsonObject obj)
{
    // 0x02 <len> bytes...
    // BER encoding. Primitive. Contents octets give the value of the
    // integer, base 256, in two's complement form, most significant
    // digit first, with the minimum number of octets. The value 0 is
    // encoded as a single 00 octet.

    byte* b;
    jsonint num = obj->x.i;
    
    assert(obj->type == JSON_INTEGER);

    if (num == 0) {
        b = allocb(3);
        b[0] = 0x02;
        b[1] = 1;
        b[2] = 0;
        obj->der_encoding_length = 3;
    } else {
        size_t nbytes;
        byte* encoded;

        encoded = encode_integer(num, &nbytes);
        b = allocb(nbytes + 2);
        b[0] = 0x02;
        b[1] = nbytes;
        memcpy(&b[2], encoded, nbytes);

        obj->der_encoding_length = nbytes+2;
    }
    obj->der_encoding = b;
}

static void set_der_encoding_real(JsonObject obj)
{
    // 0x09 <len> bytes...
    // Complicated -- see X.690 Sect. 8.5.
    // We always use base-2 encoding, meaning that we always set F=0
    byte* b = NULL;
    size_t blen = 0;            // length of the DER-encoded value
    double num = obj->x.f;
    int isnegative = signbit(num);
    int i;

    assert(obj->type == JSON_REAL);

#if DEBUG & DBG_REAL
    printf("set_der_encoding_real: num=%g (zero? %d; negative? %d)\n",
           num, (num == 0.0), isnegative);
#endif

    switch (fpclassify(num)) {
      case FP_INFINITE:
        b = allocb(3);
        b[0] = 0x09;
        b[1] = 1;
        b[2] = (isnegative ? 0x41 : 0x40);
        blen = 3;
        break;
      case FP_NAN:
        b = allocb(3);
        b[0] = 0x09;
        b[1] = 1;
        b[2] = 0x42;
        blen = 3;
        break;
      case FP_ZERO:
        if (isnegative) {
            b = allocb(3);
            b[0] = 0x09;
            b[1] = 1;
            b[2] = 0x43;
            blen = 3;
        } else {
            b = allocb(2);
            b[0] = 0x09;
            b[1] = 0;
            blen = 2;
        }
        break;
      case FP_SUBNORMAL:
        unsupported_operation("Don't currently support converting subnormal floats");
        break;
      case FP_NORMAL:
        {
            //double numnorm;
            uint64_t inorm;
            int exponent;
            size_t nbytes;
            byte* enci;

            b = allocb(2+1+2+8); /* may be an overestimate */
            b[0] = 0x09;
            blen = 3; // tag byte + length byte + first-content-octet
            
            if (isnegative) {
                num = -num;
            }

            // I want the exponent and mantissa as integers.  There's
            // a plausible-looking route to this via frexp, but that
            // extracts the mantissa as a float, which we then have to
            // convert to an integer using multiplications.
            //
            // Instead, assume 8-byte doubles, and IEEE754 floats
            // (both of which are pretty safe for normal hardware),
            // and get them by bit-twiddling.
            {
                union {
                    double d;
#if WORDS_BIGENDIAN
                    struct {
                        uint64_t s :  1;
                        uint64_t e : 11;
                        uint64_t m : 52;
                    } b;
#else
                    struct {
                        uint64_t m : 52;
                        uint64_t e : 11;
                        uint64_t s :  1;
                    } b;
#endif
                } ieee;
                uint64_t one = ((uint64_t)1)<<52;

                // XXX it would be better to check this at compile time
                // rather than run time.  However gcc seems to feel that
                // sizeof(double) is not a constant-expression, even
                // though ISO/IEC 9899:2011 6.6 para 8 suggests to me
                // that it is.
                assert(sizeof(double) == 8);

                ieee.d = num;
                inorm = ieee.b.m | one;
                exponent = ieee.b.e - 1023 - 52;
#if DEBUG & DBG_REAL
                printf("    inorm=%" PRIx64 "  exponent=%d...\n", inorm, exponent);
#endif
                // remove all of the right-most zeros, to make the
                // integer as small as possible
                while ((inorm & 0x0f) == 0) {
                    inorm >>= 4;
                    exponent += 4;
                }
                while ((inorm & 0x01) == 0) {
                    inorm >>= 1;
                    exponent += 1;
                }
            }
#if DEBUG & DBG_REAL
            printf("    .....=%" PRIx64 "  exponent=%d...\n", inorm, exponent);
#endif
            
            enci = encode_integer(exponent, &nbytes);
            assert(nbytes == 1 || nbytes == 2);

            b[2] = (isnegative ? 0xc0 : 0x80) | (nbytes == 1 ? 0x00 : 0x01);

            if (nbytes == 1) {
                b[3] = enci[0];
                blen += 1;
            } else {
                b[3] = enci[0];
                b[4] = enci[1];
                blen += 2;
            }

            enci = encode_integer(inorm, &nbytes);
            for (i=0; i<nbytes; i++, blen++) {
                b[blen] = enci[i];
            }

            b[1] = blen-2;
            
#if DEBUG & DBG_REAL
            printf("    num=%g-> inorm=%" PRIx64 " (%ld)  exponent=%d\n", num, inorm, nbytes, exponent);
#endif
        }
        break;
      default:
        fprintf(stderr, "Impossible fpclassify result?! %d\n", fpclassify(num));
        assert(0);
    }

    assert(b != NULL && blen > 0);

    obj->der_encoding = b;
    obj->der_encoding_length = blen;

#if DEBUG & DBG_REAL
    {
        printf("    --> b=");
        for (i=0; i<blen; i++) {
            printf(" %02x", b[i]);
        }
        puts("");
    }
#endif
}

static void set_der_encoding_sequence(JsonObject obj, byte tag)
{
    // tag <len> bytes...
    byte* b;
    byte* bp;
    size_t cptslen;
    size_t headlen;
    JsonObject* p = obj->x.a.p;
    size_t np = obj->x.a.len;
    int i;

    cptslen = 0;
    for (i=0; i<np; i++) {
        size_t len;
        (void) get_der_encoding_internal(p[i], &len);
        cptslen += len;
    }
    
    b = alloc_der_object(tag, cptslen, &headlen);    

    bp = &b[headlen];
    for (i=0; i<np; i++) {
        size_t len;
        byte* sub_b = get_der_encoding_internal(p[i], &len);
        memcpy(bp, sub_b, len);
        bp += len;
    }
    
    obj->der_encoding = b;
    obj->der_encoding_length = headlen + cptslen;
}

static void set_der_encoding_array(JsonObject obj) 
{
    // 0x30 <len> bytes...
    assert(obj->type == JSON_ARRAY);
    set_der_encoding_sequence(obj, 0x30);
}

/* 
 * Compare two key-value objects.
 *
 * From Kaliski, describing the DER encoding of the SET OF type:
 *
 *   DER encoding. Constructed. Contents octets are the same as for the
 *   BER encoding, except that there is an order, namely ascending
 *   lexicographic order of BER encoding. Lexicographic comparison of
 *   two different BER encodings is done as follows: Logically pad the
 *   shorter BER encoding after the last octet with dummy octets that
 *   are smaller in value than any normal octet. Scan the BER encodings
 *   from left to right until a difference is found. The smaller-valued
 *   BER encoding is the one with the smaller-valued octet at the point
 *   of difference.
 */
static int kv_compare(const void* kv1, const void* kv2)
{
    JsonObject* o1 = (JsonObject*)kv1;
    byte* b1;
    size_t b1len;
    JsonObject* o2 = (JsonObject*)kv2;
    byte* b2;
    size_t b2len;

    size_t cmplen;
    int cmp;
    
    assert((*o1)->type == JSON_KVPAIR && 
           (*o2)->type == JSON_KVPAIR);

    b1 = get_der_encoding_internal(*o1, &b1len);
    b2 = get_der_encoding_internal(*o2, &b2len);
    cmplen = (b1len < b2len ? b1len : b2len);

    cmp = memcmp(b1, b2, cmplen);
    if (cmp == 0) {
        if (b1len == b2len)
            cmp = 0;
        else if (b1len < b2len)
            cmp = -1;
        else
            cmp = +1;
    }
    
    //printf("kv_compare: %s vs %s -> %d\n", (*o1)->x.kv.k, (*o2)->x.kv.k, cmp);
    return cmp;
}

static void sort_kv_sequence(JsonObject obj)
{
    JsonObject* p;
    size_t np;

    assert(obj->type == JSON_OBJECT);

    p = obj->x.a.p;
    np = obj->x.a.len;

    if (np == 0) {
        return;                 /* JUMP OUT -- nothing to do */
    }
    
    assert(np > 0);
        
    qsort((void*)p, np, sizeof(JsonObject), kv_compare);

    return;
}

static void set_der_encoding_object(JsonObject obj)
{
    // 0x31 <len> bytes...
    assert(obj->type == JSON_OBJECT);
    sort_kv_sequence(obj);
    set_der_encoding_sequence(obj, 0x31);
}

static void set_der_encoding_kvpair(JsonObject obj)
{
    // serialised to DER as a sequence of two objects, where the
    // second may be composite
    byte* b;
    byte* bk;
    byte* bv;
    size_t bklen;
    size_t bvlen;
    size_t headlen;

    assert(obj->type == JSON_KVPAIR);
    bk = get_der_encoding_charp(obj->x.kv.k, &bklen);
    bv = get_der_encoding_internal(obj->x.kv.v, &bvlen);

    b = alloc_der_object(0x30, bklen + bvlen, &headlen);
    
    memcpy(&b[headlen], bk, bklen);
    memcpy(&b[headlen+bklen], bv, bvlen);

    obj->der_encoding = b;
    obj->der_encoding_length = headlen + bklen + bvlen;
}

static void set_der_encoding_boolean(JsonObject obj)
{
    // 0x01 0x01 0xff/0x00
    byte* b;
    assert(obj->type == JSON_BOOLEAN);

    b = allocb(3);
    b[0] = 0x01;
    b[1] = 0x01;
    b[2] = (obj->x.b ? 0xff : 0x00);

    obj->der_encoding = b;
    obj->der_encoding_length = 3;
}

static void set_der_encoding_null(JsonObject obj)
{
    // DER encoding. Primitive. Contents octets are empty; the DER
    // encoding of a NULL value is always 05 00.
    byte* b;
    assert(obj->type == JSON_NULL);
    
    b = allocb(2);
    b[0] = 0x05;
    b[1] = 0x00;
    obj->der_encoding = b;
    obj->der_encoding_length = 2;
}

static byte* get_der_encoding_internal(JsonObject obj, size_t *len)
{
    if (obj->der_encoding == NULL) {
        switch (obj->type) {
          case JSON_STRING:
            set_der_encoding_string(obj);
            break;
          case JSON_INTEGER:
            set_der_encoding_integer(obj);
            break;
          case JSON_REAL:
            set_der_encoding_real(obj);
            break;
          case JSON_ARRAY:
            set_der_encoding_array(obj);
            break;
          case JSON_OBJECT:
            set_der_encoding_object(obj);
            break;
          case JSON_KVPAIR:
            set_der_encoding_kvpair(obj);
            break;
          case JSON_BOOLEAN:
            set_der_encoding_boolean(obj);
            break;
          case JSON_NULL:
            set_der_encoding_null(obj);
            break;
          default:
            fprintf(stderr, "obj->type = %d\n", obj->type);
            assert(0);
            break;
        }
    }
    assert(obj->der_encoding != NULL);

    if (len != NULL) {
        *len = obj->der_encoding_length;
    }
    return obj->der_encoding;
}

// interface to parsing functions

// lexer functions
int yylex(void);
int lex_setup(const char*);
void lex_release(void);

// parser functions
JsonObject get_parsed_value(void);
int yyparse(void);

void yyerror(const char* msg)
{
    unsupported_operation("Parsing error: %s", msg);
}

/**
 * Parse the JSON object given in the string.  If the string cannot be
 * parsed, then this returns `NULL`, and an explanation is available from
 * the function `const char* unsupported_message()`.  The returned object
 * should be freed by @{link #free_json_object}.
 */
JsonObject parse_json_string(const char* json_string)
{
    JsonObject rval;

    assert(! in_jmp_code);

    if (setjmp(unsupported_jmp) != 0) {
        // something called unsupported_operation() (yyerror)
        rval = NULL;
    } else {
        // normal behaviour
        in_jmp_code = 1;
        unsupported_msg[0] = '\0';

        lex_setup(json_string);
        if (yyparse() == 0) {
            rval = get_parsed_value();
        } else {
            rval = NULL;
        }
        lex_release();
    }

    in_jmp_code = 0;

    return rval;
}

// DER decoding
static JsonObject parse_der_bytes_internal(byte* b, size_t blen, size_t* actuallen);


static JsonObject parse_der_ia5string_content(byte* b, size_t blen)
{
    // the content is a sequence of ASCII bytes -- simples!
    char* str;
    str = (char*)allocb(blen+1);
    memcpy(str, b, blen);
    str[blen] = '\0';
    return make_string(str);
}

static jsonint parse_der_integer_content_to_integer(byte* b, size_t blen)
{
    jsonint result;
    uintmax_t uresult;
    int i;
    jsonint one = 1;
    
    if (blen > sizeof(jsonint)) {
        unsupported_operation("DER integer too large: %ld bytes (intmax_t=%ld bytes", blen, sizeof(jsonint));
    }

    uresult = 0;
    for (i=0; i<blen; i++) {
        uresult <<= 8;
        uresult += b[i];
    }

    // uresult is a twos-complement version of the answer
    if (uresult >= one<<(8*blen-1)) {
        if (blen == sizeof(jsonint)) {
            jsonint* p;
            assert(sizeof(uintmax_t) == sizeof(jsonint));
            // exploit the machines' architecture being
            // twos-complement, and simply copy the bits
            // (attempting the 'else' calculation would overflow)
            p = (jsonint*)&uresult;
            result = *p;
        } else {
            result = - ((one<<(8*blen)) - uresult);
        }
    } else {
        result = uresult;
    }

    return result;
}

static JsonObject parse_der_integer_content(byte* b, size_t blen)
{
    jsonint result = parse_der_integer_content_to_integer(b, blen);
    return make_integer(result);
}

static JsonObject parse_der_real_content(byte* b, size_t blen)
{
    if (blen == 0) {
        return make_real(0.0);
    } else {
        int sign;
        int exponent = 0;       /* initialise to keep -Wall happy */
        int nexp;
        jsonint mantissa;
        size_t mantidx = 0;
        double result;

        if (! (b[0] & 0x80) || (b[0] & 0x3c)) {
            // must have bit 8 set, and bits 6--3 clear
            unsupported_operation("malformed first content byte for real: %02x",
                                  b[0]);
        }

        sign = (b[0] & 0x40 ? -1 : +1);
        nexp = (b[0] & 0x03);
        switch (nexp) {
          case 0:
            exponent = parse_der_integer_content_to_integer(&b[1], 1);
            mantidx = 2;
            break;
          case 1:
            exponent = parse_der_integer_content_to_integer(&b[1], 2);
            mantidx = 3;
            break;
          default:
            unsupported_operation("unexpected exponent size: %d bytes", nexp+1);
            // don't get here
        }

        assert(mantidx > 0);
        mantissa = parse_der_integer_content_to_integer(&b[mantidx], blen-mantidx);
        result = copysign(ldexp((double)mantissa, exponent), sign);

#if DEBUG & DBG_REAL
        printf("der_real: %d:%d:%ld -> %g\n", sign, exponent, mantissa, result);
#endif

        return make_real(result);
    }
}

static JsonObject parse_der_boolean_content(byte* b, size_t blen)
{
    if (blen != 1) {
        unsupported_operation("Boolean with a peculiar number of bytes (%d)", blen);
    }
    return make_boolean(b[0] ? 1 : 0);
}

static JsonObject parse_der_null_content(byte* b, size_t blen)
{
    if (blen != 0) {
        unsupported_operation("Null with content (%d bytes)", blen);
    }
    return make_null();
}

static JsonObject parse_der_sequence_content(byte* b, size_t blen)
{
    JsonObject list_of_objects = NULL;
    size_t olen;
    JsonObject obj;

    while (blen > 0) {
        obj = parse_der_bytes_internal(b, blen, &olen);
        b += olen;
        blen -= olen;
        if (list_of_objects == NULL) {
            list_of_objects = make_array(obj);
        } else {
            list_of_objects = append_object(list_of_objects, obj);
        }
    }
    return list_of_objects == NULL ? make_array(NULL) : list_of_objects;
}

static JsonObject parse_der_object_content(byte* b, size_t blen)
{
    JsonObject list_of_objects = NULL;
    size_t olen;
    JsonObject obj;
    JsonObject key = NULL;
    JsonObject value = NULL;

    // Parse each of the der objects in the content of this byte
    // array.  Each of these should be a two-element sequence (ie,
    // parsing to a JSON_ARRAY) comprising a JSON_STRING and another object.
    while (blen > 0) {
        int ok_object = 1;
        JsonObject array_obj;
#if DEBUG & DBG_SEQUENCES
        const char* pp;
#endif
   
        array_obj = parse_der_bytes_internal(b, blen, &olen);
#if DEBUG & DBG_SEQUENCES
        pp = print_json_object(array_obj);
        fprintf(stderr, "parse_der_object: obj=%s\n", pp);
        free((void*)pp);
#endif
        if (! (array_obj->type == JSON_ARRAY) && (array_obj->x.a.len == 2)) {
            ok_object = 0;
        }
        if (ok_object) {
            key = array_obj->x.a.p[0];
            if (key->type != JSON_STRING) {
                ok_object = 0;
            }
        }
        if (ok_object) {
            value = array_obj->x.a.p[1];
        }

        if (! ok_object) {
            unsupported_operation("set/object does not contain key-value pair");
        }

#if DEBUG & DBG_SEQUENCES
        pp = print_json_object(key);
        fprintf(stderr, "parse_der_object: key=%s (=%s)",
                pp, key->x.s);
        pp = print_json_object(value);
        fprintf(stderr, "; value=%s\n", pp);
#endif

        assert(ok_object && key != NULL && value != NULL);

        obj = make_kv(key->x.s, value);
        free_json_object_internal(key, 0);
        free_json_object_internal(array_obj, 0);
        
        b += olen;
        blen -= olen;
        if (list_of_objects == NULL) {
            list_of_objects = make_object(obj);
        } else {
            list_of_objects = append_object(list_of_objects, obj);
        }
    }
    return list_of_objects == NULL ? make_object(NULL) : list_of_objects;
}

static JsonObject parse_der_bytes_internal(byte* b, size_t blen, size_t* actuallen)
{
    size_t headlen;
    size_t contentlen;
    int i;
    JsonObject rval = NULL;

    if (blen < 2) {
        /* the shortest possible DER value is two bytes long
           -- this sequence is immediately wrong */
        return NULL;
    }

    if ((b[0] & 0x1f) == 0x1f) {
        unsupported_operation("Can't decode high-tag-value DER objects (found %x)", b[0]);
    }

    // determine the length of this object, and then pass the content
    // to a separate parser

    headlen = 2;
    if ((b[1] & 0x80)) {
        // long-tag-length
        int nb = (b[1] & 0x7f);
        assert(nb > 0);
        if (nb >= 8) {
            unsupported_operation("ultra-long encoded field of %x bytes (are you sure?)", nb);
        }
        headlen = 2 + nb;
        contentlen = 0;
        for (i=0; i<nb; i++) {
            contentlen = (contentlen << 8) + b[2+i];
        }
    } else {
        headlen = 2;
        contentlen = b[1];
    }
#if 0
    fprintf(stderr, "parse_der_bytes:");
    for (i=0; i<blen && i<headlen+contentlen; i++) {
        fprintf(stderr, " %02x", b[i]);
    }
    fprintf(stderr, "(%ld) -> head=%ld, content=%ld\n", blen, headlen, contentlen);
#endif

    if (headlen + contentlen > blen) {
        unsupported_operation("parse_der_bytes: tag+length is %d, but array length is %d", headlen + contentlen, blen);
    }
    
    switch(b[0]) {
      case 0x01: // boolean
        rval = parse_der_boolean_content(b+headlen, contentlen);
        break;
      case 0x02: // integer
        rval = parse_der_integer_content(b+headlen, contentlen);
        break;
      case 0x05: // null
        rval = parse_der_null_content(b+headlen, contentlen);
        break;
      case 0x09: // real
        rval = parse_der_real_content(b+headlen, contentlen);
        break;
      case 0x16: // string
        rval = parse_der_ia5string_content(b+headlen, contentlen);
        break;
      case 0x30: // sequence
        rval = parse_der_sequence_content(b+headlen, contentlen);
        break;
      case 0x31: // set of
        rval = parse_der_object_content(b+headlen, contentlen);
        break;
      default:
        unsupported_operation("Can't parse DER objects with identifier %2x\n", b[0]);
        break;
    }

    if (actuallen != NULL) {
        *actuallen = headlen + contentlen;
    }

    return rval;
}

/**
 * Decode a sequence of bytes which are the DER encoding of a JSON
 * object, and return the object.  If the byte sequence
 * cannot be parsed, then this returns `NULL`, and an explanation is
 * available from the function {@link #unsupported_message}.
 *
 * The result should be freed by
 * {@link #free_json_object} when no longer required.
 */
JsonObject parse_der_bytes(byte* b, size_t blen, size_t* actuallen)
{
    JsonObject rval = NULL;

    assert(! in_jmp_code);
    
    if (setjmp(unsupported_jmp) != 0) {
        // something has called unsupported_operation()
        rval = NULL;
    } else {
        // normal behaviour
        in_jmp_code = 1;
        unsupported_msg[0] = '\0';

        rval = parse_der_bytes_internal(b, blen, actuallen);
    }

    in_jmp_code = 0;

    return rval;
}

/**
 * Return a version string for the library.
 * The form of the string is unspecified, but is intended to be printable
 *
 * @see jason_version_number
 */
const char* jason_version_string(void)
{
    const char* snapshotid = SNAPSHOTID;
    if (snapshotid[0] == '\0') {
        // empty string -- release version
        return PACKAGE_STRING ", " RELEASE_DATE;
    } else {
        // snapshot version
        return PACKAGE_STRING ", " SNAPSHOTID;
    }
}

/**
 * Return a number which indicates the version and release numbers of
 * the library.  The number is formed by major_version * 1e6 +
 * minor_version * 1e3 + release_number
 *
 * @see jason_version_string
 */
int jason_version_number(void)
{
    return PACKAGE_VERSION_INTEGER;
}

