/*
 * This file is part of the Jason library.
 * See <https://bitbucket.org/nxg/jason>
 *
 * Copyright 2013, Norman Gray <http://nxg.me.uk>
 *
 * It is distributed under the terms of the GNU General Public
 * Licence.  See the file README.me for details.
 */

#include <stdio.h>
#include <stdlib.h> /* for exit(3) */
#include <unistd.h>
#include <assert.h>
#include <string.h>

#include "jason.h"
#include "json-parse.tab.h"

#define DTMFX_IMPLEMENTATION 1
#include "util.h"

void Usage(void);
const char* progname;

typedef enum { FMT_DER, FMT_JSON, FMT_LEXEMES } Format;


static void parse_and_display(Format input, Format output,
                              byte* input_string, size_t input_string_length,
                              FILE* out);
static void display_lexemes(const char* input);


int main(int argc, char** argv) 
{
    int ch;
    Format input_format = FMT_JSON;
    Format output_format = FMT_DER;
    byte* input_string = NULL;
    size_t input_string_length = 0;
    FILE* infile = stdin;
    FILE* outfile = stdout;

    progname = argv[0];

    while ((ch = getopt(argc, argv, "DLhi:f:o:t:V")) != -1) {
        switch (ch) {
          case 'D':             /* short for -fder -tjson */
            input_format = FMT_DER;
            output_format = FMT_JSON;
            break;
          case 'L':             /* show lexemes: for debugging */
            output_format = FMT_LEXEMES;
            break;
          case 'f':
            if (strcmp(optarg, "json") == 0) {
                input_format = FMT_JSON;
                output_format = FMT_DER;
            } else if (strcmp(optarg, "der") == 0) {
                input_format = FMT_DER;
                output_format = FMT_JSON;
            } else {
                Usage();
            }
            break;
          case 't':
            if (strcmp(optarg, "der") == 0) {
                output_format = FMT_DER;
            } else if (strcmp(optarg, "json") == 0) {
                output_format = FMT_JSON;
            } else {
                Usage();
            }
            break;
          case 'i':
            if ((infile = fopen(optarg, "r")) == NULL) {
                fprintf(stderr, "Can't open file %s to read\n", optarg);
            }
            break;
          case 'o':
            if ((outfile = fopen(optarg, "w")) == NULL) {
                fprintf(stderr, "Can't open file %s to write\n", optarg);
            }
            break;
          case 'V':
            puts(jason_version_string());
            exit(0);
            break;
          case 'h':
            /* FALL THROUGH */
          default:
            Usage();
        }
    }
    argc -= optind;
    argv += optind;

    if (argc == 0) {
#define BUFLEN 64
        byte inbuf[BUFLEN];
        StringBuilder b = make_append_buffer();

        while (! feof(infile)) {
            size_t nread = fread(inbuf, sizeof(byte), BUFLEN, infile);
            append_to_buffer_n(b, inbuf, nread);
        }
        input_string = b->buf;
        input_string_length = b->len;
#undef BUFLEN
    } else if (argc == 1) {
        if (infile != stdin) {
            fprintf(stderr, "Can't specify -i and give an argument\n");
            exit(1);
        }
        input_string = (byte*)argv[0];
        input_string_length = strlen((char*)input_string);
    } else {
        Usage();
    }

    assert(input_string != NULL);

    if ((input_format == FMT_JSON || input_format == FMT_DER)
        && (output_format == FMT_JSON || output_format == FMT_DER)) {
        parse_and_display(input_format, output_format,
                          input_string, input_string_length,
                          outfile);
    } else {
        display_lexemes((const char*)input_string);
    }

    return 0;
}

static void parse_and_display(Format input, Format output,
                              byte* input_string, size_t input_string_length,
                              FILE* out)
{
    JsonObject parse_val = NULL;

    switch (input) {
      case FMT_JSON:
        parse_val = parse_json_string((char*)input_string);
        break;
      case FMT_DER:
        parse_val = parse_der_bytes((byte*)input_string, input_string_length, NULL);
        break;
      default:
        Usage();
    }
    
    if (parse_val == NULL) {
        const char* msg = unsupported_message();
        if (msg == NULL) {
            fprintf(stderr, "Parsing failed\n");
        } else {
            fprintf(stderr, "Parsing failed: %s\n", msg);
        }
    } else {
        switch (output) {
          case FMT_JSON:
            fputs(print_json_object(parse_val), out);
            fputc('\n', out);
            break;
          case FMT_DER:
            {
                size_t blen;
                byte* b = get_der_encoding(parse_val, &blen);
                if (b == NULL) {
                    fprintf(stderr, "Failed to DER-encode that value (%s)\n",
                            unsupported_message());
                    exit(1);
                }
                fwrite(b, sizeof(byte), blen, out);
            }
            break;
          default:
            // this should be impossible
            assert(0);
        }
    }

    free_json_object(parse_val);
}

// defined in json_lex.c
int yylex(void);
char* lex_error_message(void);
int lex_setup(const char*);
void lex_release(void);

static void display_lexemes(const char* input)
{
    int l;
    
    lex_setup(input);
    
    while ((l = yylex()) != 0) {

        printf("l=%d ", l);
        
        switch (l) {
          case STRING:
            printf("string:%s\n", yylval.str);
            break;
          case INTEGER:
            printf("int:%" JSONFMT "\n", yylval.i);
            break;
          case REAL:
            printf("real:%g\n", yylval.f);
            break;
          case COLON:
            printf("colon\n");
            break;
          case COMMA:
            printf("comma\n");
            break;
          case LEFTSQ:
            printf("[\n");
            break;
          case RIGHTSQ:
            printf("]\n");
            break;
          case LEFTCURLY:
            printf("{\n");
            break;
          case RIGHTCURLY:
            printf("}\n");
            break;
          case TRUE:
            printf("TRUE\n");
            break;
          case FALSE:
            printf("FALSE\n");
            break;
          case JNULL:
            printf("NULL\n");
            break;
          default:
            printf("BAD VALUE: %d\n", l);
            break;
        }
    }

    if (lex_error_message()) {
        printf("Lexing error: %s\n", lex_error_message());
    }
    
    /* if (lex_unexpected_character() != '\0') { */
    /*     printf("Unexpected character: %c\n", lex_unexpected_character()); */
    /* } */
    
    lex_release();
}

void Usage(void)
{
    fprintf(stderr,
            "Usage: json2der [-h] [-D] [-ifile] [-ofile] [-ffmt] [-tfmt] [json-string]\n"
            "  fmt: der/json\n");
    exit(1);
}
