#include <math.h>
#include "../jason.h"
#include "lcut/lcut.h"

#define byte uint8_t


void tc_basic_json_parse(lcut_tc_t *tc, void *data)
{
    JsonObject o;
    byte* b;
    size_t blen;
    
    // Just a single uncomplicated one of each, to check the basics

    o = parse_json_string("1");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "020101", bytes_to_string(b, blen));

    o = parse_json_string("\"foo\"");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "1603666f6f", bytes_to_string(b, blen));

    o = parse_json_string("true");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "0101ff", bytes_to_string(b, blen));

    o = parse_json_string("false");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "010100", bytes_to_string(b, blen));

    o = parse_json_string("null");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "0500", bytes_to_string(b, blen));

    o = parse_json_string("[1, \"two\", true, null]");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "300d020101160374776f0101ff0500", bytes_to_string(b, blen));

    o = parse_json_string("{\"one\": 1, \"two\": 2, \"thr\": 3}");
    b = get_der_encoding(o, &blen);
    // The resulting order is important: DER-encoding requires that
    // the elements in the SET OF are ordered by the lexicographic
    // order of their contents.  The test above is chosen so that the
    // contents will have to be reordered irrespective of what order
    // they're read internally.
    LCUT_STR_EQUAL(tc,
                   "311e"
                   "300816036f6e65020101" // one:1
                   "30081603746872020103" // three:3
                   "3008160374776f020102", // two:2
                   bytes_to_string(b, blen));

    o = parse_json_string("2.0");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "0903800101", bytes_to_string(b, blen));
}

void tc_torture_syntax(lcut_tc_t *tc, void *data)
{
    JsonObject o;
    size_t blen;
    byte* b;

    o = parse_json_string("  [ \n    1   ,  \"two\"  ,\t\ttrue, null ]    \n   ");
    LCUT_ASSERT(tc, "string torture test failed", o != NULL);
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc,
                   "300d020101160374776f0101ff0500",
                   bytes_to_string(b, blen));

    // 4-byte integers
    o = parse_json_string("[16777216,16777216,16777216,16777216]");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc,
                   "3018020401000000020401000000020401000000020401000000",
                   bytes_to_string(b, blen));

    // more than 127 characters in DER encoding
    // and including a string which is 200 characters long
    o = parse_json_string("["
                          "16777216,16777216,16777216,16777216,"
                          "16777216,16777216,16777216,16777216,"
                          "16777216,16777216,16777216,16777216,"
                          "16777216,16777216,16777216,16777216,"
                          "16777216,16777216,16777216,16777216,"
                          "16777216,16777216,16777216,16777216,"
                          "\"01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\""
                          "]");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc,
                   "3082015b" // array with 347 octets of content
                   "020401000000020401000000020401000000020401000000"
                   "020401000000020401000000020401000000020401000000"
                   "020401000000020401000000020401000000020401000000"
                   "020401000000020401000000020401000000020401000000"
                   "020401000000020401000000020401000000020401000000"
                   "020401000000020401000000020401000000020401000000"
                   "1681c8" // string with 200 octets
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839"
                   "3031323334353637383930313233343536373839",
                   bytes_to_string(b, blen));

#define PARSEQ(str) do {                                                \
        o = parse_json_string(str);                                     \
        LCUT_ASSERT(tc, "parsing " str " should have succeeded", o != NULL); \
    } while (0)

    PARSEQ("{\"x\":1,\"two\":true}"); /* no spaces */
    PARSEQ("\n\t  {\t\"x\" : 1  , \n  \"two\"  :  true   }   "); /* lots of spaces everywhere */
    PARSEQ("  []");
    PARSEQ("  {}");
#undef PARSEQ
}

void tc_integers(lcut_tc_t *tc, void *data)
{
    JsonObject o;
    byte* b;
    size_t blen;

    o = parse_json_string("[0,127,128,255,+256,-127,-128,-129]");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc,
                   "301c" // array with 28 bytes of content
                   "020100"
                   "02017f"
                   "020200"
                   "80020200ff"
                   "02020100"
                   "020181"
                   "020180"
                   "0202ff7f",
                   bytes_to_string(b, blen));
    {
        // The integer conversion code depends on the machine being a
        // twos-complement machine, with sizeof(uintmax_t) being 8.
        // Check this.
        //
        // I'm pretty sure this is a direct test of whether the
        // machine is twos-complement, but I don't have a
        // non-twos-complement machine to check it on!
        unsigned char uc = 0x81;
        signed char* scp = (signed char*)&uc;
        LCUT_ASSERT(tc,
                    "This machine does not appear to be twos-complement",
                    *scp == -127);
    }

    /* integerconversion test: The following test only works if
       uintmax_t is 8 (it could be generalised, but as long as the
       test is occasionally run on a 64-bit machine, that's OK) */
    if (sizeof(uintmax_t) == 8) {
        char buf[128];
        uintmax_t one = 1;
        // presuming (u)intmax_t is 8 bytes / 64bits
        uintmax_t um = (one<<(8*sizeof(uintmax_t)-1)); // 2^31
        // sm is the largest signed integer; -sm is also a valid integer
        jsonint   sm = (jsonint)(um-1);    // needs 8 bytes to represent
        jsonint   s7 = 0x80000000000000-1; // needs 7 bytes to represent
        snprintf(buf, 128, "[ %" JSONFMT ", %" JSONFMT
                 ", %" JSONFMT ", %" JSONFMT
                 ", 32767, -32767, 127, -127 ]",
                 sm, -sm, s7, -s7);
#if 0
        fprintf(stderr,
                "uintmax_t: n=%ld um=%" PRIuMAX "/%" PRIxMAX
                "\n    sm=%" JSONFMT " -sm=%" JSONFMT "/%" PRIxMAX "\n"
                "    buf=%s\n",
                sizeof(uintmax_t), um, um, sm, -sm, -sm, buf);
#endif
        o = parse_json_string(buf);
        b = get_der_encoding(o, &blen);
#if 0
        fprintf(stderr, "actual:%s\n", bytes_to_string(b, blen));
#endif
        LCUT_STR_EQUAL(tc,
                       "3034"
                       "02087fffffffffffffff"
                       "02088000000000000001"
                       "02077fffffffffffff"
                       "020780000000000001"
                       "02027fff"
                       "02028001"
                       "02017f"
                       "020181",
                       bytes_to_string(b, blen));
    } else {
        fputs("intmax_t test skipped", stderr);
    }
}

void tc_reals(lcut_tc_t *tc, void *data)
{
    JsonObject o;
    byte* b;
    size_t blen;
    char* benc;
    char* expected;

    o = parse_json_string("[ 0.0, 1.0, -1.0, "
                          "2.0, -0.25,"
                          "1e1," // +10
                          "-1e1, -1.0e+1, -10.0, -1e+001, " // -10
                          "-1.0E-1, 0.1,"
                          "-1E12, 1.0e-2, "
                          "1361129467683753853853498429727072845824.0" // 2^130
                          " ]");

    b = get_der_encoding(o, &blen);
    benc = bytes_to_string(b, blen);
    expected = "305e"
        "0900" // 0.0
        "0903800001" // 1x2^0 = 1
        "0903c00001" // -1x2^0 = -1
        "0903800101" // 1x2^1 = 2
        "0903c0fe01" // -1x2^-2 = 0.25
        "0903800105" // +10
        "0903c00105" // -10x2^0 = -10
        "0903c00105" // ditto
        "0903c00105" // ditto
        "0903c00105" // ditto
        "0909c0c90ccccccccccccd" // -0.1 ?
        "090980c90ccccccccccccd" // +0.1
        "0906c00c0e8d4a51" // -1e+12 = 244140625 x 2^12
        "090980c5147ae147ae147b" // 1e-2 = 5764607523034235 x 2^-59
        "090481008201"; // 2^130

#if 0
    printf("--> expctd: %s\n    actual: %s\n", expected, benc);
    {
        size_t i = 0;
        
        while (expected[i] == benc[i] && expected[i] != '\0') {
            i++;
        }
        if (expected[i] == '\0' && benc[i] == '\0') {
            printf("    same\n");
        } else {
            printf("    first difference at %ld: %s != %s\n", i, &expected[i], &benc[i]);
        }
    }
#endif
    LCUT_STR_EQUAL(tc, expected, benc);

    // check encoding of special real values (these can't be read in
    // normal JSON, but it seems nice and complete to include these in
    // the translation, just in case)
    o = parse_json_string("1.0");
    o->x.f = INFINITY;
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "090140", bytes_to_string(b, blen));
    o->der_encoding = NULL;     /* hack */
    
    o->x.f = -INFINITY;
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "090141", bytes_to_string(b, blen));
    o->der_encoding = NULL;     /* hack */
    
    o->x.f = nan("");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "090142", bytes_to_string(b, blen));
    o->der_encoding = NULL;     /* hack */
    
    o->x.f = -0.0;
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc, "090143", bytes_to_string(b, blen));
    o->der_encoding = NULL;     /* hack */
    
}

void tc_objects(lcut_tc_t *tc, void *data)
{
    JsonObject o;
    size_t blen;
    byte* b;

    // Test DER-mandated sorting of the keys in an object.
    // Remember that the sort order is based on the DER-encoded
    // ["one",1], etc.
    o = parse_json_string("{"
                          "\"one\": 1,"
                          "\"two\": 2,"
                          "\"three\": 3,"
                          "\"four\": 4,"
                          "\"five\": 5,"
                          "\"six\": 6,"
                          "\"seven\": 7,"
                          "\"eight\": 8,"
                          "\"nine\": 9,"
                          "\"ten\": 10 }");
    b = get_der_encoding(o, &blen);
    LCUT_STR_EQUAL(tc,
                   "316d"
                   "300816036f6e65020101" // one
                   "30081603736978020106" // six
                   "3008160374656e02010a" // ten
                   "3008160374776f020102" // two
                   "3009160466697665020105" // five
                   "30091604666f7572020104" // four
                   "300916046e696e65020109" // nine
                   "300a16056569676874020108" // eight
                   "300a1605736576656e020107" // seven
                   "300a16057468726565020103",// three
                   bytes_to_string(b, blen));
}

void tc_strings(lcut_tc_t *tc, void *data)
{
    JsonObject o;
    byte* b;
    size_t blen;

    char* teststring = "\"q\\\" rs\\\\ s\\/ b\\b f\\f n\\n r\\r t\\t\"";
    o = parse_json_string(teststring);

    b = get_der_encoding(o, &blen);
    {
        char* expected =
            "1618"
            "712220" // q"
            "72735c20" // rs<backslash>
            "732f20" // s/
            "620820" // b<backspace>
            "660c20" // f<formfeed>
            "6e0a20" // n<newline>
            "720d20" // r<carriage-return>
            "7409";  // t<tab>
        char* actual = bytes_to_string(b, blen);
        LCUT_STR_EQUAL(tc, expected, actual);
    }

#define SHOULDFAIL(str) do {                       \
        o = parse_json_string(str);             \
        LCUT_ASSERT(tc, "parsing " str " should have failed", o == NULL); \
    } while (0)

    SHOULDFAIL("\"bogus\\x\"");
    SHOULDFAIL("\"newline\n\"");

#undef SHOULDFAIL
}

void tc_reading_der(lcut_tc_t *tc, void *data)
{
    byte der_int[]    = { 0x02, 0x01, 0x01 }; // 1
    char* str_int = "1";
    
    byte der_string[] = { 0x16, 0x03, 0x66, 0x6f, 0x6f }; // "foo"
    char* str_string = "\"foo\"";

    byte der_true[]   = { 0x01, 0x01, 0x12 }; // true (note, BER true, not DER)
    char* str_true = "true";

    byte der_false[]  = { 0x01, 0x01, 0x00 }; // false
    char* str_false = "false";

    byte der_null[]   = { 0x05, 0x00 }; // null
    char* str_null = "null";

    byte der_array[]  = { 0x30, 0x0d, // [1, "two", true, null]
                                0x02, 0x01, 0x01,
                                0x16, 0x03, 0x74, 0x77, 0x6f,
                                0x01, 0x01, 0xff,
                                0x05, 0x00 };
    char* str_array = "[ 1, \"two\", true, null ]";

    byte der_emptyarray[] = { 0x30, 0x00 };
    char* str_emptyarray = "[  ]";

    byte der_emptyobject[] = { 0x31, 0x00 };
    char* str_emptyobject = "{  }";
    
    byte der_object[] = { 0x31, 0x14, // {"one": 1, "two": true}
                                0x30, 0x08,
                                      0x16, 0x03, 0x6f, 0x6e, 0x65,
                                      0x02, 0x01, 0x01,
                                0x30, 0x08,
                                      0x16, 0x03, 0x74, 0x77, 0x6f,
                                      0x01, 0x01, 0x01 }; // BER true
    char* str_object = "{ \"one\": 1, \"two\": true }";

    byte der_floatarray[] = { 0x30, 0x3f,  // 63
                              0x09, 0x00,  // 0.0
                              0x09, 0x03, 0x80, 0x00, 0x01,  // 1x2^0 = 1
                              0x09, 0x03, 0xc0, 0x00, 0x01,  // -1x2^0 = -1
                              0x09, 0x03, 0x80, 0x01, 0x01,  // 1x2^1 = 2
                              0x09, 0x03, 0xc0, 0xfe, 0x01,  // -1x2^-2 = -0.25, 
                              0x09, 0x03, 0xc0, 0x01, 0x05,  // -10x2^0 = -10
                              0x09, 0x09, 0x80, 0xc9, 0x0c, 0xcc, 0xcc, 0xcc, 0xcc, 0xcc, 0xcd,  // +0.1
                              0x09, 0x06, 0xc0, 0x0c, 0x0e, 0x8d, 0x4a, 0x51,  // -1e-12 = 244140625 x 2^12
                              0x09, 0x09, 0x80, 0xc5, 0x14, 0x7a, 0xe1, 0x47, 0xae, 0x14, 0x7b,  // 1e-2 = 5764607523034235 x 2^-59
                              0x09, 0x04, 0x81, 0x00, 0x82, 0x01 } ; // 2^130
    char* str_floatarray = 
        "[ 0.0, 1.0, -1.0, 2.0, -0.25, -10.0, 0.1, -1000000000000.0, 0.01, 1361129467683753853853498429727072845824.0 ]";

    byte der_string_with_escapes[] = { 0x16, 0x18, 
                                       0x71, 0x22, 0x20, 
                                       0x72, 0x73, 0x5c, 0x20, 
                                       0x73, 0x2f, 0x20, 
                                       0x62, 0x08, 0x20, 
                                       0x66, 0x0c, 0x20, 
                                       0x6e, 0x0a, 0x20, 
                                       0x72, 0x0d, 0x20, 
                                       0x74, 0x09 };
    char* str_string_with_escapes =
        "\"q\\\" rs\\\\ s\\/ b\\b f\\f n\\n r\\r t\\t\"";
    
    const char* json_s;

#define TESTDECODE(arr, str) do {                  \
        json_s = print_json_object(parse_der_bytes(arr, sizeof(arr), NULL)); \
        LCUT_STR_EQUAL(tc, str, json_s);                                \
        if (json_s != NULL) {                                           \
            free((void*)json_s);                                        \
        }                                                               \
      } while (0)

    TESTDECODE(der_string, str_string);
    TESTDECODE(der_int, str_int);
    TESTDECODE(der_true, str_true);
    TESTDECODE(der_false, str_false);
    TESTDECODE(der_null, str_null);

    // the following tests are slightly sensitive to the (irrelevant)
    // details of the prettyprinting of JSON
    TESTDECODE(der_array, str_array);
    TESTDECODE(der_emptyarray, str_emptyarray);
    TESTDECODE(der_emptyobject, str_emptyobject);
    
    if (sizeof(uintmax_t) == 8) {
        // see the integerconversion test above
        // (try this one only if sizeof(jsonint)==8
        byte der_intarray[]
            = { 0x30, 0x34, 
                0x02, 0x08, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
                0x02, 0x08, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 
                0x02, 0x07, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
                0x02, 0x07, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 
                0x02, 0x02, 0x7f, 0xff, 
                0x02, 0x02, 0x80, 0x01, 
                0x02, 0x01, 0x7f, 
                0x02, 0x01, 0x81 };
    
        char buf[128];
        uintmax_t one = 1;
        // presuming (u)intmax_t is 8 bytes / 64bits
        uintmax_t um = (one<<(8*sizeof(uintmax_t)-1)); // 2^31
        // sm is the largest signed integer; -sm is also a valid integer
        jsonint   sm = (jsonint)(um-1);    // needs 8 bytes to represent
        jsonint   s7 = 0x80000000000000-1; // needs 7 bytes to represent
        snprintf(buf, 128, "[ %" JSONFMT ", %" JSONFMT
                 ", %" JSONFMT ", %" JSONFMT
                 ", 32767, -32767, 127, -127 ]",
                 sm, -sm, s7, -s7);
        json_s = print_json_object(parse_der_bytes(der_intarray,
                                                   sizeof(der_intarray),
                                                   NULL));
        LCUT_STR_EQUAL(tc, buf, json_s);
        if (json_s != NULL) {
            free((void*)json_s);
        }
    }

    TESTDECODE(der_object, str_object);
    
    TESTDECODE(der_floatarray, str_floatarray);
    TESTDECODE(der_string_with_escapes, str_string_with_escapes);

#undef TESTDECODE
}


int main() {
    lcut_ts_t   *suite = NULL;

    LCUT_TEST_BEGIN("dtmfx tests", NULL, NULL);

    LCUT_TS_INIT(suite, "dtmfx tests", NULL, NULL);
    LCUT_TC_ADD(suite, "Basic json_parse", tc_basic_json_parse, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Integer encoding", tc_integers, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Real encoding", tc_reals, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "String encoding", tc_strings, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Syntax torture", tc_torture_syntax, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Object wrinkles", tc_objects, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Reading DER",  tc_reading_der, NULL, NULL, NULL);
    LCUT_TS_ADD(suite);

    LCUT_TEST_RUN();
    LCUT_TEST_REPORT();
    LCUT_TEST_END();

    LCUT_TEST_RESULT();
}
